#!/usr/bin/env python3

"""
List of available functions:

"""

# Standard imports
import re
from io import TextIOWrapper
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, List, Union, Optional, Any, Tuple
from datetime import datetime
from urllib.parse import urlparse, urlunparse

import requests
# Libica imports
from libica.openapi.v2 import ApiClient, ApiException
from libica.openapi.v2.api.project_data_api import ProjectDataApi
from libica.openapi.v2.api.project_data_copy_batch_api import ProjectDataCopyBatchApi
from libica.openapi.v2.model.analysis_input_external_data import AnalysisInputExternalData
from libica.openapi.v2.model.aws_temp_credentials import AwsTempCredentials
from libica.openapi.v2.model.create_data import CreateData
from libica.openapi.v2.model.create_project_data_copy_batch import CreateProjectDataCopyBatch
from libica.openapi.v2.model.create_project_data_copy_batch_item import CreateProjectDataCopyBatchItem
from libica.openapi.v2.model.create_temporary_credentials import CreateTemporaryCredentials
from libica.openapi.v2.model.data_id_or_path_list import DataIdOrPathList
from libica.openapi.v2.model.data_url_with_path import DataUrlWithPath
from libica.openapi.v2.model.download import Download
from libica.openapi.v2.model.job import Job
from libica.openapi.v2.model.project_data import ProjectData
from libica.openapi.v2.model.project_data_copy_batch import ProjectDataCopyBatch
from libica.openapi.v2.model.temp_credentials import TempCredentials
from libica.openapi.v2.model.upload import Upload

# Local imports
from ...enums import DataType, ProjectDataSortParameters, ProjectDataStatusValues
from ...utils.configuration import get_icav2_configuration, logger
from ...utils.globals import LIBICAV2_DEFAULT_PAGE_SIZE
from ...utils.miscell import is_uuid_format


def get_project_data_file_id_from_project_id_and_path(
        project_id: str,
        file_path: Path,
        create_file_if_not_found: bool = False
) -> str:
    """
    Given a project id, parent folder path and file_name, return the file id
    If the file is not found, and create_file_if_not_found is True, create the file

    :param project_id:  The project id
    :param file_path:  The path to the file
    :param create_file_if_not_found:  Create the file object if it does not exist

    :return: The file id

    :raises: FileNotFoundError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import (
            create_download_urls, get_project_folder_id_from_project_id_and_path
        )
        from wrapica.libica_models import DataUrlWithPath

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        file_id: str = get_project_file_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            file_path=Path("/path/to/file.txt")
        )

        download_urls: List[DataUrlWithPath] = create_download_urls(
            project_id="proj.abcdef1234567890",
            folder_id=project_folder_obj.data.id,
            recursive=True
        )

        for download_url in download_urls:
            print(download_url.url)

    """
    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(file_path.parent.absolute()) + "/"
    if parent_folder_path == "//":
        parent_folder_path = "/"

    # Add the filename to the list of filenames to search on
    filename = [
        file_path.name
    ]

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the list of project data.
        data_items: List[ProjectData] = api_instance.get_project_data_list(
            project_id=project_id,
            parent_folder_path=parent_folder_path,
            filename=filename,
            filename_match_mode="EXACT",
            file_path_match_mode="FULL_CASE_INSENSITIVE",
            type="FILE"
        ).items
    except ApiException as e:
        if not create_file_if_not_found:
            logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
            raise ApiException
        else:
            file_obj = create_file_in_project(
                project_id=project_id,
                file_path=file_path.absolute(),
            )
            return file_obj.data.id

    # Get the file id
    try:
        file_id = next(
            filter(
                lambda data_iter: data_iter.data.details.path == str(file_path),
                data_items
            )
        )
    except StopIteration:
        logger.error("Could not find file id for file: %s\n" % file_path)
        raise FileNotFoundError

    return file_id.data.id


def create_data_in_project(
        project_id: str,
        parent_folder_path: Path,
        data_name: str,
        data_type: DataType
) -> ProjectData:
    """
    Create a data object in a project context

    :param project_id: The project ID
    :param parent_folder_path: The parent folder path of where the data object needs to be created
    :param data_name:  The name of the file or folder
    :param data_type:  One of DataType.FILE or DataType.FOLDER

    :return: The newly created project data object
    :rtype: List[`ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import (
            create_data_in_project, create_data_in_project,
        )
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = create_data_in_project(
            project_id="abcd-1234-efab-5678",
            parent_folder_path=Path("/path/to/folder/"),
            data_name="file.txt",
            data_type=DataType.FILE
        )
    """

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(parent_folder_path.absolute()) + "/"
    if parent_folder_path == "//":
        parent_folder_path = "/"

    # example passing only required values which don't have defaults set
    try:
        # Create a project data.
        api_response: ProjectData = api_instance.create_data_in_project(
            project_id=project_id,
            create_data=CreateData(
                name=data_name,
                folder_path=parent_folder_path,
                data_type=DataType(data_type).value,
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_project_data: %s\n" % e)
        raise ApiException

    # Return the folder id
    return api_response


def create_file_in_project(
        project_id: str,
        file_path: Path,
) -> ProjectData:
    """
    Create a file in a project

    :param project_id: The project id to create the file in
    :param file_path: The path to the file

    :return: The newly created file

    :rtype: `ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import create_file_in_project
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = create_file_in_project(
            project_id="abcd-1234-efab-5678",
            file_path=Path("/path/to/file.txt")
        )
    """
    return create_data_in_project(
        project_id=project_id,
        parent_folder_path=file_path.parent,
        data_name=file_path.name,
        data_type=DataType.FILE
    )


def create_folder_in_project(
        project_id: str,
        folder_path: Path,
) -> ProjectData:
    """
    Create a folder in a project

    :param project_id:  The project ID
    :param folder_path:  The path to the folder to create

    :return: The newly created folder project data object
    :rtype: `ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import create_folder_in_project
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = create_folder_in_project(
            project_id="abcd-1234-efab-5678",
            folder_path=Path("/path/to/folder/new/")
        )

    """
    return create_data_in_project(
        project_id=project_id,
        parent_folder_path=folder_path.parent,
        data_name=folder_path.name,
        data_type=DataType.FOLDER
    )


def get_project_data_folder_id_from_project_id_and_path(
        project_id: str,
        folder_path: Path,
        create_folder_if_not_found: bool = False
) -> str:
    """
    Given a project_id and a path, return the folder_id.

    Note that given the folder_path is a Path object (which do not end in /),
    we need to append a '/' to the end of the path before calling the API
    In the case that the folder_path is the root folder, we ensure that only a single '/' is provided.

    :param project_id:  The project id to search in
    :param folder_path:  The path to the folder
    :param create_folder_if_not_found:  If folder does not exist in project, do we want to create it?

    :return: The folder id

    :raises: NotADirectoryError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_folder_id_from_project_id_and_path

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        folder_id: str = get_project_folder_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            folder_path=Path("/path/to/folder/")
        )
    """
    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(folder_path.parent.absolute()) + "/"
    # Exception for when folder is in the top directory
    if parent_folder_path == '//':
        parent_folder_path = '/'

    # Add the folder name to the list of folder names to search on
    folder_name = [
        folder_path.name
    ]

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the list of project data.
        data_items: List[ProjectData] = api_instance.get_project_data_list(
            project_id=project_id,
            parent_folder_path=parent_folder_path,
            filename=folder_name,
            filename_match_mode="EXACT",
            file_path_match_mode="FULL_CASE_INSENSITIVE",
            type="FOLDER"
        ).items
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
        raise ApiException

    # Get the folder id
    try:
        folder_id: ProjectData = next(
            filter(
                lambda data_iter: data_iter.data.details.path == str(folder_path) + "/",
                data_items
            )
        )
    except StopIteration:
        if create_folder_if_not_found:
            # Create the folder
            folder_id = create_folder_in_project(
                project_id=project_id,
                folder_path=folder_path
            )
        else:
            logger.error("Could not find folder id for folder: %s\n" % folder_path)
            raise NotADirectoryError

    return folder_id.data.id


def get_project_data_id_from_project_id_and_path(
        project_id: str,
        data_path: Path,
        data_type: DataType,
        create_data_if_not_found: bool = False
) -> str:
    """
    Given a project_id and a path, return the data_id, where DATA_TYPE is one of FILE or FOLDER
    Should call the underlying get_data_id_from_project_id_and_path function split by data type

    :param project_id: The project context the data exists in
    :param data_path: The path to the data in the project
    :param data_type: The data_type, one of DataType.FILE, DataType.FOLDER
    :param create_data_if_not_found:

    :raises: FileNotFoundError, NotADirectoryError, ApiException

    :return: The data id

    :note:
      Use get_file_id_from_project_id_and_path or get_folder_id_from_project_id_and_path instead if data_type is known

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_data_id_from_project_id_and_path
        from wrapica.enums import DataType

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        # Get a folder
        data_id: str = get_project_data_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/folder/"),
            data_type=DataType.FOLDER
        )

        # Get a file
        data_id: str = get_project_data_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/file.txt"),
            data_type=DataType.FILE
        )
    """
    if data_type == DataType.FOLDER:
        return get_project_data_folder_id_from_project_id_and_path(
            project_id=project_id,
            folder_path=data_path,
            create_folder_if_not_found=create_data_if_not_found
        )
    else:
        return get_project_data_file_id_from_project_id_and_path(
            project_id=project_id,
            file_path=data_path,
            create_file_if_not_found=create_data_if_not_found
        )


def get_project_data_obj_by_id(
        project_id: str,
        data_id: str
) -> ProjectData:
    """
    Given a project_id and a data_id, return the data object

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: The project data object
    :rtype: `ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import get_project_data_obj_by_id
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = get_project_data_obj_by_id(
            project_id="abcd-1234-efab-5678",
            data_id="fil.abcdef1234567890"
        )
    """

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the list of project data.
        data_obj: ProjectData = api_instance.get_project_data(
            project_id=project_id,
            data_id=data_id
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
        raise ApiException

    return data_obj


def get_project_data_obj_from_project_id_and_path(
        project_id: str,
        data_path: Path,
        data_type: DataType,
        create_data_if_not_found: bool = False
) -> ProjectData:
    """
    Given a project_id and a path, return the data object, where DATA_TYPE is one of FILE or FOLDER
    Will call the get_project_data_id and then call get_project_data_obj_by_id

    :param project_id: The project id to search in
    :param data_path: The path to the data in the project
    :param data_type: The data_type, one of DataType.FILE, DataType.FOLDER
    :param create_data_if_not_found: If the data is not found, and create_data_if_not_found is True, create the data

    :return: The project data object
    :rtype: `ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_

    :raises: FileNotFoundError, NotADirectoryError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_data_obj_from_project_id_and_path
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        # Get a folder project data object
        project_folder_data_obj: ProjectData = get_project_data_obj_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/folder/"),
            data_type=DataType.FOLDER
        )

        # Get a file project data object
        project_file_data_obj: ProjectData = get_project_data_obj_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/file.txt"),
            data_type=DataType.FILE
        )

        print(project_folder_data_obj.data.id)
        # fol.abcdef1234567890

        print(project_file_data_obj.data.id)
        # fil.abcdef1234567890
    """
    # Collect the data id, either fol.id or fil.id
    project_data_id: str = get_project_data_id_from_project_id_and_path(
        project_id=project_id,
        data_path=data_path,
        data_type=data_type,
        create_data_if_not_found=create_data_if_not_found
    )

    # Then collect the object itself
    return get_project_data_obj_by_id(
        project_id=project_id,
        data_id=project_data_id
    )


def get_project_data_path_by_id(
        project_id: str,
        data_id: str
) -> Path:
    """
    Given a project id and data id, return the path of the data

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: The path of the data
    :rtype: Path

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_data_path_by_id

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_path: Path = get_project_data_path_by_id(
            project_id="abcd-1234-efab-5678",
            data_id="fil.abcdef1234567890"
        )

        print(project_data_path)
        # /path/to/file.txt
    """
    project_data_path = get_project_data_obj_by_id(
        project_id=project_id,
        data_id=data_id
    ).data.details.path

    return Path(project_data_path)


def list_project_data_non_recursively(
        project_id: str,
        parent_folder_id: Optional[str] = None,
        parent_folder_path: Optional[Path] = None,
        file_name: Optional[Union[str, List[str]]] = None,
        status: Optional[Union[ProjectDataStatusValues, List[ProjectDataStatusValues]]] = None,
        data_type: Optional[DataType] = None,
        creation_date_after: Optional[datetime] = None,
        creation_date_before: Optional[datetime] = None,
        status_date_after: Optional[datetime] = None,
        status_date_before: Optional[datetime] = None,
        sort: Optional[Union[ProjectDataSortParameters, List[ProjectDataSortParameters]]] = ""
) -> List[ProjectData]:
    """
    Given a project id and parent folder id or path,
    return a list of data objects that are directly under that folder

    :param project_id: The project id to search in
    :param parent_folder_path: The path to the parent folder (can use parent_folder_id instead)
    :param parent_folder_id: The parent folder id (can use parent_folder_path instead)
    :param file_name: The name of the file or directory to look for, can also be a list of names, may also use * as a wildcard
    :param status: The status of the data, one of ProjectDataStatusValues
    :param data_type: The type of the data, one of DataType.FILE, DataType.FOLDER
    :param creation_date_after: Return only data created after this date
    :param creation_date_before: Return only data created before this date
    :param status_date_after: Return only data with status date after this date
    :param status_date_before: Return only data with status date before this date
    :param sort: The sort order, one or more of ProjectDataSortParameters (Use '-' prefix to sort in descending order)
      * timeCreated - Sort by time created
      * timeModified - Sort by time modified
      * name - Sort by name
      * path - Sort by path
      * fileSizeInBytes - Sort by file size in bytes
      * status - Sort by status
      * format - Sort by format
      * dataType - Sort by data type
      * willBeArchivedAt - Sort by when the data will be archived
      * willBeDeletedAt - Sort by when the data will be deleted

    :return: List of data objects
    :rtype: List[`ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_]

    :raises: AssertionError, ApiException, ValueError

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import list_project_data_non_recursively
        from wrapica.libica_models import ProjectData, ProjectDataSortParameters
        from wrapica.enums import ProjectDataStatusValues, DataType

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = list_project_data_non_recursively(
            project_id="abcd-1234-efab-5678",
            parent_folder_path=Path("/path/to/folder/"),
            file_name="file.txt",
            status=ProjectDataStatusValues.COMPLETED,
            data_type=DataType.FILE,
            creation_date_after=datetime(2021, 1, 1),
            creation_date_before=datetime(2021, 12, 31),
            status_date_after=datetime(2021, 1, 1),
            status_date_before=datetime(2021, 12, 31),
            sort=ProjectDataSortParameters.TIME_CREATED
        )

        for project_data in project_data_list:
            print(project_data.data.details.name)

    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if parent_folder_id is None and parent_folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif parent_folder_id is not None and parent_folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError
    # Specify either parent_folder_id as a list parent_folder_path is just as a string
    if parent_folder_id is not None:
        parent_folder_id = [parent_folder_id]

    # Convert parent folder path to a string
    if parent_folder_path is not None:
        parent_folder_path = str(parent_folder_path.absolute()) + "/"
        if parent_folder_path == "//":
            parent_folder_path = "/"

    # Check file_name
    if isinstance(file_name, str):
        file_name = [file_name]

    # Check status
    if status is not None:
        if isinstance(status, ProjectDataStatusValues):
            status = [status]
        elif isinstance(status, str):
            status = [ProjectDataStatusValues(status)]
        status = list(
            map(
                lambda status_iter: status_iter.value,
                status
            )
        )

    # Check data_type
    if data_type is not None:
        if isinstance(data_type, (DataType, str)):
            data_type = DataType(data_type).value

    # Check sort
    if sort is not None:
        if isinstance(sort, ProjectDataSortParameters):
            sort = [sort]
        elif isinstance(sort, str):
            sort = [ProjectDataSortParameters(sort)]
        # Complete a comma join of the sort parameters
        sort = ", ".join(
            list(
                map(
                    lambda sort_iter: sort_iter.value,
                    sort
                )
            )
        )

    # Collect api instance
    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectDataApi(api_client)

    # Set other parameters
    page_size = LIBICAV2_DEFAULT_PAGE_SIZE
    page_token = ""
    # We use page tokens if sort is None, otherwise we use page offsets
    if sort is not None:
        page_offset = 0
    else:
        page_offset = ""

    # Initialise data ids - we may need to extend the items multiple times
    data_ids: List[ProjectData] = []

    # Loop through the pages
    while True:
        # Attempt to collect all data ids
        try:
            # Retrieve the list of project data
            api_response = api_instance.get_project_data_list(
                project_id=project_id,
                parent_folder_ids=parent_folder_id,
                parent_folder_path=parent_folder_path,
                page_size=str(page_size),
                page_offset=str(page_offset),
                page_token=page_token,
                file_name=file_name,
                status=status,
                data_type=data_type,
                creation_date_after=creation_date_after,
                creation_date_before=creation_date_before,
                status_date_after=status_date_after,
                status_date_before=status_date_before,
                sort=sort
            )
        except ApiException as e:
            raise ValueError("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)

        # Extend items list
        data_ids.extend(api_response.items)

        if sort is not None:
            # Check page offset and page size against total item count
            if page_offset + page_size > api_response.total_item_count:
                break
            page_offset += page_size
        else:
            # Check if there is a next page
            if api_response.next_page_token is None:
                break
            page_token = api_response.next_page_token

    return data_ids


def find_project_data_recursively(
        project_id: str,
        parent_folder_id: str,
        parent_folder_path: Path,
        name: str,
        data_type: DataType,
        min_depth: Optional[int] = None,
        max_depth: Optional[int] = None
) -> List[ProjectData]:
    """
    Given a project_id, a parent_folder_id, a data_name and a data_type, return a list of data objects
    This is a slow exercise and should only be used if the max_depth is low and the total number of items in the
    directory is very high

    :param project_id: The project id to search in
    :param parent_folder_id: The parent folder id (alternative to parent_folder_path)
    :param parent_folder_path: The path to the parent folder (alternative to parent_folder_id)
    :param name: The name of the file or directory to look for, may also use * as a wildcard
    :param data_type: The type of the data, one of DataType.FILE, DataType.FOLDER
    :param min_depth: The minimum depth to search
    :param max_depth: The maximum depth to search

    :return: List of data objects
    :rtype: List[`ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_]

    :raises: ApiException, AssertionError, ValueError

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import find_project_data_recursively
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = find_project_data_recursively(
            project_id="abcd-1234-efab-5678",
            parent_folder_id="fol.abcdef1234567890",
            name="file.txt",
            data_type=DataType.FILE,
            min_depth=1,
            max_depth=3
        )

        for project_data in project_data_list:
            print(project_data.data.details.name)

    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if parent_folder_id is None and parent_folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif parent_folder_id is not None and parent_folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError

    # Matched data items thing we return
    matched_data_items: List[ProjectData] = []
    name_regex_obj = re.compile(name)

    # Get top level items
    data_items: List[ProjectData] = list_project_data_non_recursively(
        project_id=project_id,
        parent_folder_id=parent_folder_id,
        parent_folder_path=parent_folder_path,
        data_type=data_type
    )

    # Check if we can pull out any items in the top directory
    if min_depth is None or min_depth <= 1:
        for data_item in data_items:
            data_item_match = name_regex_obj.match(data_item.data.details.name)
            if data_type is not None and not data_item.data.details.data_type == data_type:
                continue
            if data_item_match is not None:
                matched_data_items.append(data_item)

    # Otherwise look recursively
    if max_depth is None or not max_depth <= 1:
        # Listing sub folders
        subfolders = list(
            filter(
                lambda x: x.data.details.data_type == "FOLDER",
                data_items
            )
        )
        for subfolder in subfolders:
            matched_data_items.extend(
                find_project_data_recursively(
                    project_id=project_id,
                    parent_folder_id=subfolder.data.id,
                    parent_folder_path=subfolder.data.details.path,
                    name=name,
                    data_type=data_type,
                    min_depth=min_depth - 1 if min_depth is not None else None,
                    max_depth=max_depth - 1 if max_depth is not None else None
                )
            )

        return matched_data_items


def find_project_data_bulk(
        project_id: str,
        parent_folder_id: Optional[str] = None,
        parent_folder_path: Optional[Path] = None,
        data_type: Optional[DataType] = None
) -> List[ProjectData]:
    """
    Given a project_id and a parent_folder_id, return a list of all data objects in the folder (recursively)

    :param project_id: The project id to search in
    :param parent_folder_id: The parent folder id (alternative to parent_folder_path)
    :param parent_folder_path: The path to the parent folder (alternative to parent_folder_id)
    :param data_type: The type of the data, one of DataType.FILE, DataType.FOLDER

    :return: List of data objects
    :rtype: List[`ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_]

    :raises: ApiException, AssertionError

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import find_project_data_bulk
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = find_project_data_bulk(
            project_id="abcd-1234-efab-5678",
            parent_folder_id="fol.abcdef1234567890",
            data_type=DataType.FILE
        )

        for project_data in project_data_list:
            print(project_data.data.details.name)
    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if parent_folder_id is None and parent_folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif parent_folder_id is not None and parent_folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError

    # Get the parent folder path as a string
    if parent_folder_path is None:
        parent_folder_path = str(get_project_data_path_by_id(project_id, parent_folder_id)) + "/"
    else:
        parent_folder_path = str(parent_folder_path.absolute()) + "/"

    # Initialise
    data_ids: List[ProjectData] = []
    # Collect api instance
    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectDataApi(api_client)

    # Set other parameters
    page_size = LIBICAV2_DEFAULT_PAGE_SIZE
    page_token = ""

    # Iterate over all pages
    while True:
        # Attempt to collect all data ids
        try:
            # Retrieve the list of project data
            api_response = api_instance.get_project_data_list(
                project_id=project_id,
                file_path=[parent_folder_path],
                file_path_match_mode="STARTS_WITH_CASE_INSENSITIVE",
                page_size=str(page_size),
                page_token=page_token,
                type=data_type.value
            )

        except ApiException as e:
            logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
            raise ApiException

        # Extend items list
        data_ids.extend(api_response.items)

        # Check page offset and page size against total item count
        page_token = api_response.next_page_token

        if page_token == "":
            break

    return data_ids


def create_download_url(
        project_id: str,
        file_id: str
) -> str:
    """
    Given a project_id and a data_id, create a presigned url for a file
    in project-pipeline, or view a file in project-data
    Create download URL

    :param project_id: The owning project id
    :param file_id: The id of the file

    :return: The download URL string
    :rtype: str

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import create_download_url

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        download_url: str = create_download_url(
            project_id="proj.abcdef1234567890",
            file_id="fil.abcdef1234567890"
        )

        print(download_url)
        # https://s3.amazonaws.com/umccr-illumina-prod/abcd-1234-efab-5678/abcdef1234567890

    """
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

        # example passing only required values which don't have defaults set
        try:
            # Retrieve a download URL for this data.
            api_response: Download = api_instance.create_download_url_for_data(
                project_id,
                file_id
            )
        except ApiException as e:
            logger.error("Exception when calling ProjectDataApi->create_download_url_for_data: %s\n" % e)
            raise ApiException

    return api_response.get("url")


def create_download_urls(
        project_id: str,
        folder_id: str,
        recursive: bool = False
) -> List[DataUrlWithPath]:
    """
    Given a project data folder return a list where each item is an object with the following attributes

    :param project_id: The owning project id
    :param folder_id:  The id of the folder
    :param recursive:  Whether to provide download urls recursively

    :return: List of download urls
    :rtype: List[`DataUrlWithPath <https://umccr-illumina.github.io/libica/openapi/v2/docs/DataUrlWithPathList/>`_]

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import (
            create_download_urls, get_project_folder_id_from_project_id_and_path
        )
        from wrapica.libica_models import ProjectData, DataUrlWithPath

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_folder_obj: ProjectData = get_project_folder_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            folder_path=Path("/path/to/folder/")
        )

        download_urls: List[DataUrlWithPath] = create_download_urls(
            project_id="proj.abcdef1234567890",
            folder_id=project_folder_obj.data.id,
            recursive=True
        )

        for download_url in download_urls:
            print(download_url.url)
    """

    if recursive:
        project_data_list = find_project_data_bulk(
            project_id=project_id,
            parent_folder_id=folder_id,
            data_type=DataType.FILE
        )
    else:
        project_data_list = list_project_data_non_recursively(
            project_id=project_id,
            parent_folder_id=folder_id,
            data_type=DataType.FILE
        )

    # Set data paths
    data_id_paths_list = DataIdOrPathList(
        data_ids=list(
            map(
                lambda project_file_iter: project_file_iter.data.id,
                project_data_list
            )
        )
    )

    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve download URLs for the data.
        api_response = api_instance.create_download_urls_for_data(project_id, data_id_paths_list)
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_download_urls_for_data: %s\n" % e)
        raise ApiException

    # Return items
    return api_response.items


def convert_icav2_uri_to_data_obj(
        data_uri: str,
        create_data_if_not_found: bool = False
) -> ProjectData:
    """
    Given an ICAv2 URI, return a project data object

    :param data_uri: The icav2 uri to convert to a data object
    :param create_data_if_not_found:  If the data is not found, and create_data_if_not_found is True, create the data

    :return: libica v2 Project Data Object
    :rtype: `Project Data <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_

    :raises: ValueError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import convert_icav2_uri_to_data_obj, ProjectData

        project_data_object: ProjectData = convert_icav2_uri_to_data_obj(
            "icav2://project-name/path/to/data/"
        )

        print(project_data_object.data.id)
        # file.abcdef1234567890
    """
    # Import other functions locally to avoid circular imports
    from ...project import get_project_id_from_project_name

    data_uri_obj = urlparse(data_uri)

    # Set data type
    if data_uri_obj.path.endswith("/"):
        data_type = DataType.FOLDER
    else:
        data_type = DataType.FILE

    # Check if the project is in project id format
    if is_uuid_format(data_uri_obj.netloc):
        project_id = data_uri_obj.netloc
    else:
        project_id = get_project_id_from_project_name(data_uri_obj.netloc)

    # Return the data object
    return get_project_data_obj_from_project_id_and_path(
        project_id=project_id,
        data_path=Path(data_uri_obj.path),
        data_type=data_type,
        create_data_if_not_found=create_data_if_not_found
    )


def convert_project_data_obj_to_icav2_uri(
    project_data: ProjectData
) -> str:
    """
    Return the file object as an icav2:// uri

    :param project_data: The ProjectData object

    :return: The icav2:// uri string
    """
    return str(
        urlunparse((
            "icav2",
            project_data.project_id,
            project_data.data.details.path + ("/" if project_data.data.details.data_type == "FOLDER" else ""),
            None, None, None
        ))
    )


def convert_project_id_and_data_path_to_icav2_uri(
    project_id: str,
    data_path: Path,
    data_type: DataType
) -> str:
    """
    Given a project_id and a data_id, return the icav2:// uri

    Unlike convert_project_data_obj_to_icav2_uri, this does not require the path to exist.

    If the data_type is DataType.FOLDER, the path should end with a forward slash.

    :param project_id: The project id to search in
    :param data_path: The path to the data in the project
    :param data_type: The data_type, one of DataType.FILE, DataType.FOLDER

    :return: The icav2:// uri string
    :rtype: str

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_data import convert_project_id_and_data_path_to_icav2_uri
        from wrapica.enums import DataType

        icav2_uri: str = convert_project_id_and_data_path_to_icav2_uri(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/folder/"),
            data_type=DataType.FOLDER
        )
    """

    return str(
        urlunparse((
            "icav2",
            project_id,
            str(data_path) + ("/" if data_type == DataType.FOLDER else ""),
            None, None, None
        ))
    )


def get_aws_credentials_access_for_project_folder(
        project_id: str,
        folder_id: Optional[str],
        folder_path: Optional[Path]
) -> AwsTempCredentials:
    """
    Given a project_id and a folder_id or folder_path, collect the AWS Access Credentials for downloading this data.

    :param project_id: The project id of the data
    :param folder_id: The folder id (alternative to folder_path)
    :param folder_path: The folder path (alternative to folder_id)

    :return: An object with the following attributes:
      * access_key
      * secret_key
      * session_token
      * region
      * bucket
      * object_prefix
      * server_side_encryption_algorithm
      * server_side_encryption_key

    :rtype: `AwsTempCredentials <https://umccr-illumina.github.io/libica/openapi/v2/docs/AwsTempCredentials/>`_

    :raises: AssertionError, ApiException, ValueError

    :Examples:

    .. code-block:: python
        :linenos:

        import subprocess
        from wrapica.project_data import get_aws_credentials_access_for_project_folder
        from wrapica.libica_models import AwsTempCredentials

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        aws_temp_credentials: AwsTempCredentials = get_aws_credentials_access_for_project_folder(
            project_id="proj.abcdef1234567890",
            folder_path=Path("/path/to/folder/")
        )

        local_path = Path("/path/to/local/folder/")

        subprocess.run(
          [
            "aws", "s3", "sync",
            # Can add sync parameters here like --dryrun or --exclude / --include
            "s3://{}/{}".format(aws_temp_credentials.bucket, aws_temp_credentials.object_prefix),
            str(local_path)
          ],
          env={
            "AWS_ACCESS_KEY_ID": aws_temp_credentials.access_key,
            "AWS_SECRET_ACCESS_KEY": aws_temp_credentials.secret_key,
            "AWS_SESSION_TOKEN": aws_temp_credentials.session_token,
            "AWS_REGION": aws_temp_credentials.region
          }
        )
    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if folder_id is None and folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif folder_id is not None and folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError

    if folder_id is None:
        folder_id = get_project_data_folder_id_from_project_id_and_path(
            project_id=project_id,
            folder_path=folder_path
        )

    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    create_temporary_credentials = CreateTemporaryCredentials()

    # example passing only required values which don't have defaults set
    try:
        # Retrieve temporary credentials for this data.
        api_response: TempCredentials = api_instance.create_temporary_credentials_for_data(
            project_id, folder_id, create_temporary_credentials=create_temporary_credentials
        )
    except ApiException as e:
        logger.warning("Exception when calling ProjectDataApi->create_temporary_credentials_for_data: %s\n" % e)
        raise ValueError

    return api_response.aws_temp_credentials


def is_folder_id_format(folder_id_str: str) -> bool:
    """
    Does this string look like a folder id?

    :param folder_id_str: The string to check

    :return: True if the string looks like a folder id
    :rtype: bool

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import is_folder_id_format

        print(is_folder_id_format("fol.abcdef1234567890"))
        # True
    """
    return re.match("fol.[0-9a-f]{32}", folder_id_str) is not None


def is_file_id_format(file_id_str: str) -> bool:
    """
    Does this string look like a folder id?

    :param file_id_str: The string to check

    :return: True if the string looks like a file id
    :rtype: bool

    :Examples:

    .. code-block:: python
    :linenos:

        from wrapica.project_data import is_file_id_format

        print(is_file_id_format("fil.abcdef1234567890"))
        # True

    """
    return re.match("fil.[0-9a-f]{32}", file_id_str) is not None


def is_data_id(data_id: str) -> bool:
    """
    Check if data id is a data id

    :param data_id: The data id to check

    :return: True if the data id is a data id

    :rtype: bool

    :Examples:

    .. code-block:: python

    from wrapica.project_data import is_data_id

    print(is_data_id("fil.abcdef1234567890"))
    # True
    """
    return is_file_id_format(data_id) or is_folder_id_format(data_id)


def presign_cwl_directory(
        project_id: str,
        data_id: str
) -> List[
    Union[
        Dict[str, Union[Union[dict, str], Any]],
        Dict[str, Union[str, Any]]
    ]
]:
    """
    Given a CWL directory object, presign all files in the directory recursively, and return the list of presigned url

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: The CWL input json Directory object where each file in the listing has a presigned url for a location attributes
    :rtype: List[Dict[str, Union[Union[dict, str], Any]]]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import presign_cwl_directory

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id
        # Use wrapica.project_data.get_folder_id_from_project_id_and_path
        # If you need to convert a folder path to a folder id

        cwl_directory: List[Dict[str, Union[Union[dict, str], Any]]] = presign_cwl_directory(
            project_id="abcd-1234-efab-5678",
            data_id="fol.abcdef1234567890"
        )

        print(cwl_directory)
        # [
        #   {
        #     "class": "Directory",
        #     "basename": "folder",
        #     "listing": [
        #       {
        #         "class": "File",
        #         "basename": "file.txt",
        #         "location": "https://s3.amazonaws.com/umccr-illumina-prod/abcd-1234-efab-5678/abcdef1234567890"
        #       }
        #     ]
        #   }
        # ]
    """
    # Data ids
    cwl_item_objs = []

    # List items noncursively
    file_obj_list = list_project_data_non_recursively(
        project_id=project_id,
        parent_folder_id=data_id
    )

    # Collect file object list
    for file_item_obj in file_obj_list:
        data_type: str = file_item_obj.get("data").get("details").get('data_type')  # One of FILE | FOLDER
        data_id = file_item_obj.get("data").get("id")
        basename = file_item_obj.get("data").get("details").get("name")
        if data_type == "FOLDER":
            cwl_item_objs.append(
                {
                    "class": "Directory",
                    "basename": basename,
                    "listing": presign_cwl_directory(project_id, data_id)
                }
            )
        else:
            cwl_item_objs.append(
                {
                    "class": "File",
                    "basename": basename,
                    "location": create_download_url(project_id, data_id)
                }
            )

    return cwl_item_objs


def presign_cwl_directory_with_external_data_mounts(
    project_id: str,
    data_id: str
) -> Tuple[
    # External data mounts
    List[AnalysisInputExternalData],
        # Dict listing
    List[Dict]
]:
    """
    Given a cwl directory with a listing attribute, presign all files in the directory recursively, and return the
    list of presigned url mount objects and the cwl directory listing object

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: external_data_mounts, cwl_item_objs
    :rtype: Tuple[List[`AnalysisInputExternalData <https://umccr-illumina.github.io/libica/openapi/v2/docs/AnalysisInputExternalData/>`_], List[Dict]]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import presign_cwl_directory_with_external_data_mounts

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id
        # Use wrapica.project_data.get_folder_id_from_project_id_and_path
        # If you need to convert a folder path to a folder id

        external_data_mounts, cwl_item_objs = presign_cwl_directory_with_external_data_mounts(
            project_id="abcd-1234-efab-5678",
            data_id="fol.abcdef1234567890"
        )

        print(external_data_mounts)
        # [
        #   {
        #     "url": "https://s3.amazonaws.com/umccr-illumina-prod/abcd-1234-efab-5678/abcdef1234567890",
        #     "type": "FILE",
        #     "mount_path": "abcd-1234-efab-5678/abcdef1234567890/file.txt"
        #   }
        # ]

        print(cwl_item_objs)
        # [
        #   {
        #     "class": "Directory",
        #     "basename": "folder",
        #     "listing": [
        #       {
        #         "class": "File",
        #         "basename": "file.txt",
        #         "location": "abcd-1234-efab-5678/abcdef1234567890/file.txt"
        #       }
        #     ]
        #   }
        # ]

    """
    # Data ids
    cwl_item_objs = []

    # External data mounts
    external_data_mounts = []

    # List items noncursively
    file_obj_list = list_project_data_non_recursively(
        project_id=project_id,
        parent_folder_id=data_id
    )

    # Collect file object list
    for file_item_obj in file_obj_list:
        data_type: str = file_item_obj.get("data").get("details").get('data_type')  # One of FILE | FOLDER
        data_id = file_item_obj.get("data").get("id")
        basename = file_item_obj.get("data").get("details").get("name")
        if data_type == "FOLDER":
            listing, external_data_mounts_new = presign_cwl_directory_with_external_data_mounts(
                project_id,
                data_id
            )
            external_data_mounts.extend(external_data_mounts_new)
            cwl_item_objs.append(
                {
                    "class": "Directory",
                    "basename": basename,
                    "listing": listing
                }
            )
        else:
            # Generate presigned url
            presigned_url = create_download_url(project_id, data_id)

            # Generate mount path for file
            mount_path = str(
                Path(project_id) /
                Path(data_id) /
                Path(basename)
            )

            # Append the mount path and presigned url to the external data mounts list
            external_data_mounts.append(
                AnalysisInputExternalData(
                    url=presigned_url,
                    type=DataType.FILE.value,
                    mount_path=mount_path
                )
            )

            # Append the item to the cwl item object list
            cwl_item_objs.append(
                {
                    "class": "File",
                    "basename": basename,
                    "location": mount_path
                }
            )

    return external_data_mounts, cwl_item_objs


def read_icav2_file_contents(
    project_id: str,
    data_id: str,
    output_path: Union[Path, TextIOWrapper]
):
    """
    Write icav2 file contents to a path

    :param project_id:
    :param data_id:
    :param output_path:
    :return:
    """
    if isinstance(output_path, Path):
        # Ensure parent directory exists
        if not output_path.parent.exists():
            logger.error(f"Could not write to output path {output_path} as the parent directory does not exist")
            raise NotADirectoryError

    # Get the presigned url
    presigned_url = create_download_url(project_id, data_id)

    # Get the file contents with the requests package
    r = requests.get(presigned_url)

    if isinstance(output_path, Path):
        # Write the file contents to the output path
        with open(output_path, "wb") as f:
            f.write(r.content)
    else:
        # Write the file contents to the output path
        output_path.write(r.content.decode())


def read_icav2_file_contents_to_string(
    project_id: str,
    data_id: str
) -> str:
    """

    Stream down the icav2 file contents and return as a string

    :param project_id: The project id
    :param data_id: The data id

    :return: The file contents as a string
    :rtype: str

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_data import read_icav2_file_contents_to_string

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        file_contents: str = read_icav2_file_contents_to_string(
            project_id="abcd-1234-efab-5678",
            data_id="fil.abcdef1234567890"
        )

        print(file_contents)
        # this is the file contents
    """

    with NamedTemporaryFile() as temp_file_h:
        read_icav2_file_contents(
            project_id=project_id,
            data_id=data_id,
            output_path=Path(temp_file_h.name)
        )

        with open(temp_file_h.name, "r") as f:
            return f.read()


def get_project_data_upload_url(
    project_id: str,
    data_id: str
):
    # The client must configure the authentication and authorization parameters
    # in accordance with the API server security policy.

    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve an upload URL for this data.
        api_response: Upload = api_instance.create_upload_url_for_data(project_id, data_id)
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_upload_url_for_data: %s\n" % e)
        raise ApiException

    return api_response.url


def write_icav2_file_contents(
    project_id: str,
    data_path: Path,
    file_stream_or_path: Union[Path, TextIOWrapper]
):
    """
    Write icav2 file contents to a path

    :param project_id:
    :param data_path:
    :param file_stream_or_path:
    :return:
    """

    # Generate a new file in the project
    new_file_obj = create_file_in_project(
        project_id=project_id,
        file_path=data_path
    )

    # Get the upload url
    upload_url = get_project_data_upload_url(
        project_id=project_id,
        data_id=new_file_obj.data.id
    )

    if isinstance(file_stream_or_path, Path):
        with open(file_stream_or_path, "rb") as f:
            file_contents = f.read()
    else:
        file_contents = file_stream_or_path.read()

    # Upload file contents with the requests package
    requests.put(upload_url, data=file_contents)

    # Return the new file id
    return new_file_obj.data.id


def get_file_by_file_name_from_project_data_list(
    file_name: str,
    project_data_list: List[ProjectData]
) -> ProjectData:
    """
    Useful for collecting a file object from an analysis output object

    :param file_name: The name of the file to get
    :param project_data_list: The list of project data objects to search through

    :return: The file object
    :rtype: `ProjectData <https://umccr-illumina.github.io/libica/openapi/v2/docs/ProjectData/>`_

    :raises: ValueError

    :Examples:

    .. code-block:: python

        from wrapica.project_data import get_file_by_file_name_from_project_data_list

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = find_project_data_bulk(
            project_id="abcd-1234-efab-5678",
            parent_folder_id="fol.abcdef1234567890",
            data_type=DataType.FILE
        )

        file_obj: ProjectData = get_file_by_file_name_from_project_data_list(
            file_name="file.txt",
            project_data_list=project_data_list
        )
    """

    # Find the first file with this name
    try:
        return next(
            filter(
                lambda file_iter: (
                    file_iter.data.details.name == file_name and
                    file_iter.data.details.data_type == DataType.FILE.value
                ),
                project_data_list
            )
        )
    except StopIteration:
        logger.error(f"Could not get file {file_name} from analysis output")
        raise ValueError


def project_data_copy_batch_handler(
    source_data_ids: List[str],
    destination_project_id: str,
    destination_folder_path: Path
) -> Job:
    """
    Copy a batch of files from one project to another
    """

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataCopyBatchApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Copy a batch of project data.
        api_response: ProjectDataCopyBatch = api_instance.create_project_data_copy_batch(
            project_id=destination_project_id,
            create_project_data_copy_batch=CreateProjectDataCopyBatch(
                items=list(
                    map(
                        lambda source_data_id_iter: CreateProjectDataCopyBatchItem(
                            data_id=source_data_id_iter
                        ),
                        source_data_ids
                    )
                ),
                destination_folder_id=get_project_data_folder_id_from_project_id_and_path(
                    project_id=destination_project_id,
                    folder_path=destination_folder_path,
                    create_folder_if_not_found=True
                ),
                copy_user_tags=True,
                copy_technical_tags=True,
                copy_instrument_info=True,
                action_on_exist="SKIP"
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->copy_project_data_batch: %s\n" % e)
        raise ApiException

    # Return the job ID for the project data copy batch
    return api_response.job



#!/usr/bin/env python3

"""
Useful functions for working with wss:// protocol
"""

# External imports
import websocket
from websocket._exceptions import WebSocketTimeoutException, WebSocketBadStatusException
from pathlib import Path
from bs4 import BeautifulSoup

# Local imports
from .logger import get_logger

# Set logger
logger = get_logger()


def write_websocket_to_file(url: str, output_file: Path):
    ws = websocket.WebSocket()
    try:
        ws.connect(url, timeout=3)
    except WebSocketBadStatusException:
        logger.error(f"Couldn't connect to websocket url {url}, try again in a few moments"
                     f"Websocket has likely closed and log is being written to a file")

    with open(output_file, "w") as output_h:
        while True:
            try:
                line = ws.recv_data()[-1].decode()
                output_h.write(line)
            except WebSocketTimeoutException:
                logger.info("Got to the end of the websocket, or the connection timed out?")
                break


def convert_html_to_text(input_file: Path, output_file: Path):
    with open(input_file, "r") as input_h, open(output_file, "w") as output_h:
        output_h.write(BeautifulSoup(input_h, features="lxml").get_text())



#!/usr/bin/env python3

"""
Collect logger through verboselogs package
"""

# External imports
import verboselogs
import logging
import inspect

# Globals
LOGGER_STYLE = "%(asctime)s - %(levelname)-8s - %(module)-25s - %(funcName)-40s : LineNo. %(lineno)-4d - %(message)s"

# Set logger
logger = verboselogs.VerboseLogger(__name__)
logger.addHandler(logging.StreamHandler())


def get_caller_function():
    """
    Get the function that was used to call the previous
    Some loggers report <module>
    :return:
    """
    # Get the inspect stack trace
    inspect_stack = inspect.stack()

    # Since we're already in a function, we need the third attribute
    # i.e function of interest -> function that called this one -> this function
    frame_info = inspect_stack[2]

    # Required attribute is' function
    function_id = getattr(frame_info, "function", None)

    if function_id is None:
        # Don't really want to break on this just yet but code is ready to go for it.
        return None
    else:
        return function_id


def set_basic_logger():
    """
    Set the basic logger before we then take in the --deploy-env values to see where we write to
    :return:
    """
    # Get a basic logger
    logger = logging.getLogger()

    # Get a stderr handler
    console = logging.StreamHandler()

    # Set level
    console.setLevel(logging.INFO)

    # Set format
    formatter = logging.Formatter(LOGGER_STYLE)
    console.setFormatter(formatter)

    logger.addHandler(console)

    return logger


def set_logger(script_dir, script, deploy_env, log_level=logging.INFO):
    """
    Initialise a logger
    :return:
    """
    new_logger = logging.getLogger()
    new_logger.setLevel(log_level)

    # create a logging format
    formatter = logging.Formatter(LOGGER_STYLE)

    # create a file handler
    #file_handler = RotatingFileHandler(filename=script_dir / "{}.{}".format(script, LOG_FILE_SUFFIX[deploy_env]),
    #                                   maxBytes=100000000, backupCount=5)
    ## Set Level
    #file_handler.setLevel(log_level)
    #file_handler.setFormatter(formatter)

    # create a console handler
    console_handler = logging.StreamHandler()
    # Hard coded as don't need too much verbosity on the console side
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)

    # add the handlers to the logger
    #new_logger.addHandler(file_handler)
    new_logger.addHandler(console_handler)


def get_logger():
    """
    Return logger object
    :return:
    """
    function_that_called_this_one = get_caller_function()
    return logging.getLogger(function_that_called_this_one)





#!/usr/bin/env python

# Import everything
from .functions.project_functions import (
    # Project Data functions
    get_project_id_from_project_name
)

__all__ = [
    # Functions
    'get_project_id_from_project_name'
]



#!/usr/bin/env python3

"""
Functions that don't quite do anywhere else
"""
# Standard imports
import re
from typing import Dict, Any, Union, List
from urllib.parse import urlparse
from uuid import UUID


def camel_to_snake_case(camel_case: str) -> str:
    return re.sub(r'(?<!^)(?=[A-Z])', '_', camel_case).lower()


def snake_to_camel_case(snake_case: str) -> str:
    return ''.join(word.title() for word in snake_case.split('_'))


def to_lower_camel_case(snake_str):
    # We capitalize the first letter of each component except the first one
    # with the 'capitalize' method and join them together.
    camel_string = snake_to_camel_case(snake_str)
    return snake_str[0].lower() + camel_string[1:]


def sanitise_dict_keys(input_dict: Dict) -> Dict:
    output_dict = {}
    key: str
    value: Any
    for key, value in input_dict.items():
        output_dict[camel_to_snake_case(key)] = value
    return output_dict


def is_uuid_format(project_id: str) -> bool:
    try:
        _ = UUID(project_id, version=4)
        return True
    except ValueError:
        return False


def is_uri_format(uri_str: str) -> bool:
    """
    Determine if the string is a valid URI
    :param uri_str:
    :return:
    """
    try:
        _ = urlparse(uri_str)
        return True
    except ValueError:
        return False


def build_curl_body_from_libica_item(libica_item: Any) -> Union[Dict, Any]:
    """
    Useful for pipeline retention, i.e what was actually submitted to the orchestration engine
    :param libica_item:
    :return:
    """
    if not isinstance(libica_item, object) or isinstance(libica_item, str):
        return libica_item
    open_api_body_dict = {}
    for key, value in libica_item._data_store.items():
        if isinstance(value, List):
            output_value = [
                build_curl_body_from_libica_item(value_item)
                for value_item in value
            ]
        elif isinstance(value, object) and hasattr(value, "_data_store"):
            output_value = build_curl_body_from_libica_item(value)
        else:
            output_value = value
        open_api_body_dict[libica_item.attribute_map.get(key)] = output_value
    return open_api_body_dict



#!/usr/bin/env python

# Import everything
from .functions.project_analyses_functions import (
    # Project Analysis functions
    get_project_analysis_inputs,
    get_analysis_input_object_from_analysis_code,
    get_outputs_object_from_analysis_id,
    get_analysis_output_object_from_analysis_code,
    get_analysis_obj,
    get_analysis_steps,
    get_analysis_log_from_analysis_step,
    write_analysis_step_logs
)

__all__ = [
    # Functions
    'get_project_analysis_inputs',
    'get_analysis_input_object_from_analysis_code',
    'get_outputs_object_from_analysis_id',
    'get_analysis_output_object_from_analysis_code',
    'get_analysis_obj',
    'get_analysis_steps',
    'get_analysis_log_from_analysis_step',
    'write_analysis_step_logs'
]



#!/usr/bin/env python

"""
Models to easily import
"""

from libica.openapi.v2 import ApiClient, ApiException
from libica.openapi.v2.api.analysis_storage_api import AnalysisStorageApi
from libica.openapi.v2.api.entitlement_detail_api import EntitlementDetailApi
from libica.openapi.v2.api.project_analysis_api import ProjectAnalysisApi
from libica.openapi.v2.api.project_pipeline_api import ProjectPipelineApi

from libica.openapi.v2.model.activation_code_detail import ActivationCodeDetail
from libica.openapi.v2.model.analysis import Analysis
from libica.openapi.v2.model.analysis import Analysis
from libica.openapi.v2.model.analysis_input import AnalysisInput
from libica.openapi.v2.model.analysis_output import AnalysisOutput
from libica.openapi.v2.model.analysis_data_input import AnalysisDataInput
from libica.openapi.v2.model.analysis_input_data_mount import AnalysisInputDataMount
from libica.openapi.v2.model.analysis_input_external_data import AnalysisInputExternalData
from libica.openapi.v2.model.analysis_output_mapping import AnalysisOutputMapping
from libica.openapi.v2.model.analysis_parameter_input import AnalysisParameterInput
from libica.openapi.v2.model.analysis_step import AnalysisStep
from libica.openapi.v2.model.analysis_step_logs import AnalysisStepLogs
from libica.openapi.v2.model.analysis_tag import AnalysisTag
from libica.openapi.v2.model.aws_temp_credentials import AwsTempCredentials
from libica.openapi.v2.model.create_cwl_analysis import CreateCwlAnalysis
from libica.openapi.v2.model.create_data import CreateData
from libica.openapi.v2.model.create_nextflow_analysis import CreateNextflowAnalysis
from libica.openapi.v2.model.create_temporary_credentials import CreateTemporaryCredentials
from libica.openapi.v2.model.cwl_analysis_input import CwlAnalysisInput
from libica.openapi.v2.model.cwl_analysis_json_input import CwlAnalysisJsonInput
from libica.openapi.v2.model.cwl_analysis_structured_input import CwlAnalysisStructuredInput
from libica.openapi.v2.model.data_id_or_path_list import DataIdOrPathList
from libica.openapi.v2.model.data_url_with_path import DataUrlWithPath
from libica.openapi.v2.model.download import Download
from libica.openapi.v2.model.input_parameter import InputParameter
from libica.openapi.v2.model.input_parameter_list import InputParameterList
from libica.openapi.v2.model.nextflow_analysis_input import NextflowAnalysisInput
from libica.openapi.v2.model.pipeline_configuration_parameter import PipelineConfigurationParameter
from libica.openapi.v2.model.pipeline_configuration_parameter_list import PipelineConfigurationParameterList
from libica.openapi.v2.model.project_data import ProjectData
from libica.openapi.v2.model.project_paged_list import ProjectPagedList
from libica.openapi.v2.model.project_pipeline import ProjectPipeline
from libica.openapi.v2.model.project_pipeline_list import ProjectPipelineList
from libica.openapi.v2.model.search_matching_activation_codes_for_cwl_analysis import SearchMatchingActivationCodesForCwlAnalysis
from libica.openapi.v2.model.search_matching_activation_codes_for_nextflow_analysis import SearchMatchingActivationCodesForNextflowAnalysis
from libica.openapi.v2.model.temp_credentials import TempCredentials
from libica.openapi.v2.model.job import Job



#!/usr/bin/env python3

# External imports
import subprocess

# Local imports
from .logger import get_logger

# Get logger
logger = get_logger()


def run_subprocess_proc(*args, **kwargs):
    """
    Utilities runner for running a subprocess command and printing log files
    :param args:
    :param kwargs:
    :return:
    """

    subprocess_proc = subprocess.run(*args, **kwargs)

    command_str = '" "'.join(map(str, ["\'%s\'" % arg for arg in subprocess_proc.args])) \
        if type(subprocess_proc.args) == list \
        else subprocess_proc.args
    command_str = '"' + command_str + '"'

    # Get outputs
    if subprocess_proc.stdout is not None:
        command_stdout = subprocess_proc.stdout.decode()
    else:
        command_stdout = None

    if subprocess_proc.stderr is not None:
        command_stderr = subprocess_proc.stderr.decode()
    else:
        command_stderr = None

    # Get return code
    command_returncode = subprocess_proc.returncode

    if not command_returncode == 0:
        # Print returncode to warning
        logger.debug("Received exit code \"{}\" for command {}".format(
            command_returncode,
            command_str
        ))
        # Print stdout/stderr to console
        if command_stdout is not None:
            logger.debug("Stdout was: \n\"{}\"".format(command_stdout.strip()))
        if command_stderr is not None:
            logger.debug("Stderr was: \n\"{}\"".format(command_stderr.strip()))
    else:
        # Let debug know command returned successfully
        logger.debug("Command \"{}\" returned successfully".format(command_str))
        # Print stdout/stderr to console
        if command_stdout is not None:
            logger.debug("Stdout was: \n\"{}\"".format(command_stdout.strip()))
        if command_stderr is not None:
            logger.debug("Stderr was: \n\"{}\"".format(command_stderr.strip()))

    return command_returncode, command_stdout, command_stderr



#!/usr/bin/env python

# Standard imports
import json
from io import TextIOWrapper
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Union, Dict, Any, Optional

# Libica imports
from libica.openapi.v2 import ApiClient, ApiException
from libica.openapi.v2.api.project_analysis_api import ProjectAnalysisApi
from libica.openapi.v2.model.analysis import Analysis
from libica.openapi.v2.model.analysis_input import AnalysisInput
from libica.openapi.v2.model.analysis_output import AnalysisOutput
from libica.openapi.v2.model.analysis_output_list import AnalysisOutputList
from libica.openapi.v2.model.analysis_step import AnalysisStep
from libica.openapi.v2.model.analysis_step_logs import AnalysisStepLogs
from libica.openapi.v2.model.cwl_analysis_output_json import CwlAnalysisOutputJson
from libica.openapi.v2.api import project_analysis_api

# Util imports
from ...utils.configuration import get_icav2_configuration
from ...utils.websocket_helpers import write_websocket_to_file, convert_html_to_text
from ...utils.logger import get_logger

logger = get_logger()


def get_project_analysis_inputs(
    project_id: str,
    analysis_id: str
) -> List[AnalysisInput]:
    """
    Get the analysis inputs for a given analysis

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: List of analysis inputs
    :rtype: List[`AnalysisInput <https://umccr-illumina.github.io/libica/openapi/v2/docs/AnalysisInput>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_project_analysis_inputs

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        workflow_inputs = get_project_analysis_inputs(project_id, analysis_id)

    """

    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

        # example passing only required values which don't have defaults set
        try:
            # Retrieve the outputs of an analysis.
            analysis_input_list: List[AnalysisInput] = api_instance.get_analysis_inputs(
                project_id, analysis_id
            ).items
        except ApiException as e:
            logger.error("Exception when calling ProjectAnalysisApi->get_analysis_outputs: %s\n" % e)
            raise ApiException

    return analysis_input_list


def get_analysis_input_object_from_analysis_code(
    project_id: str,
    analysis_id: str,
    analysis_code: str
) -> AnalysisInput:
    """
    Given an analysis code for an analysis id, collect the analysis input object
    Confirm the input has either analysis or external data attributes

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query
    :param analysis_code: The analysis input code to query

    :return: The analysis input object
    :rtype: `AnalysisInput <https://umccr-illumina.github.io/libica/openapi/v2/docs/AnalysisInput>`_

    :raises: StopIteration, ValueError, ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_input_object_from_analysis_code

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"
        analysis_code = "run_folder"

        run_folder_input_data_id = get_analysis_input_object_from_analysis_code(
            project_id, analysis_id, analysis_code
        ).analysis_data[0].data_id
    """

    # Get analysis inputs
    analysis_input_list: List[AnalysisInput] = get_project_analysis_inputs(
        project_id,
        analysis_id
    )

    # Iterate through inputs to find the one we want
    try:
        input_obj: AnalysisInput = next(
            filter(
                lambda analysis_iter: analysis_iter.code == analysis_code,
                analysis_input_list
            )
        )
    except StopIteration:
        logger.error(f"Could not get {analysis_code} from analysis {analysis_id}")
        raise StopIteration

    if len(input_obj.analysis_data) == 0 and len(input_obj.external_data) == 0:
        logger.error(f"Expected analysis data or external data to be 1 but got {len(input_obj.analysis_data)}")
        raise ValueError

    return input_obj


def get_outputs_object_from_analysis_id(
    project_id: str,
    analysis_id: str
) -> List[AnalysisOutput]:
    """
    Query the outputs object from the analysis id

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: List of analysis outputs
    :rtype: List[`AnalysisOutput <https://umccr-illumina.github.io/libica/openapi/v2/docs/AnalysisOutput>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
    
        :linenos:
        from wrapica.project_analysis import get_outputs_object_from_analysis_id

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        workflow_outputs = get_outputs_object_from_analysis_id(project_id, analysis_id)
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = project_analysis_api.ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the outputs of an analysis
        api_response: AnalysisOutputList = api_instance.get_analysis_outputs(project_id, analysis_id)
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis_outputs: %s\n" % e)
        raise ApiException

    return api_response.items


def get_analysis_output_object_from_analysis_code(
    project_id: str,
    analysis_id: str,
    analysis_code: str
) -> AnalysisOutput:
    """
    Given an analysis code for an analysis id, collect the analysis output object

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query
    :param analysis_code: The analysis output code to collect

    :return: The analysis output object
    :rtype: `AnalysisOutput <https://umccr-illumina.github.io/libica/openapi/v2/docs/AnalysisOutput>`_

    :raises: StopIteration, ValueError, ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_output_object_from_analysis_code

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"
        analysis_code = "Output"

        analysis_folder_output_id = get_analysis_output_object_from_analysis_code(
            project_id, analysis_id, analysis_code
        ).data[0].data_id
    """

    analysis_output: List[AnalysisOutput] = get_outputs_object_from_analysis_id(
        project_id,
        analysis_id
    )
    try:
        output_obj: AnalysisOutput = next(
            filter(
                lambda analysis_iter: analysis_iter.code == analysis_code,
                analysis_output
            )
        )
    except StopIteration:
        logger.error(f"Could not get output item from analysis {analysis_id}")
        raise StopIteration

    if len(output_obj.data) == 0:
        logger.error(f"Expected analysis output data to be at least 1 but got {len(output_obj.data)}")
        raise ValueError

    return output_obj


def get_cwl_outputs_json_from_analysis_id(
    project_id: str,
    analysis_id: str
) -> Dict[str, Any]:
    """
    Query the outputs object from the analysis id

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: List of analysis outputs
    :rtype: Dict[str, Any]

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_cwl_outputs_json_from_analysis_id

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        cwl_json_output = get_cwl_outputs_json_from_analysis_id(project_id, analysis_id)
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = project_analysis_api.ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the outputs of an analysis
        api_response: CwlAnalysisOutputJson = api_instance.get_cwl_output_json(
            project_id, analysis_id
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis_outputs: %s\n" % e)
        raise ApiException

    return json.loads(api_response.output_json)


def get_analysis_obj(
    project_id: str,
    analysis_id: str
) -> Analysis:
    """
    Get an analysis object given a project id and analysis id

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: The analysis object
    :rtype: `Analysis <https://umccr-illumina.github.io/libica/openapi/v2/docs/Analysis>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_obj

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        analysis = get_analysis_obj(project_id, analysis_id)
    """

    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = project_analysis_api.ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve an analysis.
        api_response: Analysis = api_instance.get_analysis(project_id, analysis_id)
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis: %s\n" % e)
        raise ApiException

    return api_response


def get_analysis_steps(project_id: str, analysis_id: str, include_technical_steps: bool = False) -> List[AnalysisStep]:
    """
    Get the workflow steps for a given analysis

    :param project_id:
    :param analysis_id:

    :return: List of analysis steps
    :rtype: List[`AnalysisStep <https://umccr-illumina.github.io/libica/openapi/v2/docs/AnalysisStep>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_steps

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        analysis_step_list = get_analysis_steps(project_id, analysis_id)
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = project_analysis_api.ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the individual steps of an analysis.
        api_response = api_instance.get_analysis_steps(
            project_id,
            analysis_id
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis_steps: %s\n" % e)
        raise ApiException

    # Collect all steps
    analysis_steps: List[AnalysisStep] = api_response.items

    if not include_technical_steps:
        analysis_steps = list(
            filter(
                lambda step_iter: step_iter.technical is False,
                analysis_steps
            )
        )

    return analysis_steps


def get_analysis_log_from_analysis_step(analysis_step: AnalysisStep) -> AnalysisStepLogs:
    """
    Get the logs for a given analysis step

    :param analysis_step:

    :return: Get the logs attribute of an analysis step
    :rtype: `AnalysisStepLogs <https://umccr-illumina.github.io/libica/openapi/v2/docs/AnalysisStepLogs>`_

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_steps

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        analysis_step_list = get_analysis_steps(project_id, analysis_id)

        step_logs = list(map(get_analysis_log_from_analysis_step, analysis_step_list))
    """
    return analysis_step.logs


def write_analysis_step_logs(
    project_id: str,
    step_logs: AnalysisStepLogs,
    log_name: str,
    output_path: Union[Path | TextIOWrapper],
    is_cwltool_log: Optional[bool] = False
) -> None:
    """
    Write the analysis step logs to a file

    :param project_id: The project id the analysis was run in
    :param step_logs: The step logs object
    :param log_name: One of stdout or stderr
    :param output_path: The output path to write the file to

    :raises: ApiException, NotADirectoryError

    :return: None
    :rtype: None

    :Examples:

    .. code-block:: python

        :linenos:
        from pathlib import Path
        from wrapica.project_analysis import get_analysis_steps

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        # Get analysis steps and logs
        analysis_step_list = get_analysis_steps(project_id, analysis_id)
        step_logs = list(map(get_analysis_log_from_analysis_step, analysis_step_list))

        # Write first step logs to file
        write_analysis_step_logs(
            project_id=project_id,
            step_logs=step_logs[0],
            log_name="stderr",
            output_path=Path("stderr.log")
        )

    """
    # Local imports
    from ...project_data import read_icav2_file_contents

    # Check if we're getting our log from a stream
    is_stream = False
    log_stream = None
    log_data_id = ""

    # Initialise list of non empty log attributes
    non_empty_log_attrs = []

    # Check attributes of log obj
    for attr in dir(step_logs):
        if attr.startswith('_'):
            continue
        if getattr(step_logs, attr) is None:
            continue
        non_empty_log_attrs.append(attr)

    if log_name == "stdout":
        if hasattr(step_logs, "std_out_stream") and step_logs.std_out_stream is not None:
            is_stream = True
            log_stream = step_logs.std_out_stream
        elif hasattr(step_logs, "std_out_data") and step_logs.std_out_data is not None:
            log_data_id: str = step_logs.std_out_data.id
        else:
            logger.error("Could not get either file output or stream of logs")
            logger.error(f"The available attributes were {', '.join(non_empty_log_attrs)}")
            raise AttributeError
    else:
        if hasattr(step_logs, "std_err_stream") and step_logs.std_err_stream is not None:
            is_stream = True
            log_stream = step_logs.std_err_stream
        elif hasattr(step_logs, "std_err_data") and step_logs.std_err_data is not None:
            log_data_id: str = step_logs.std_err_data.id
        else:
            logger.error("Could not get either file output or stream of logs")
            logger.error(f"The available attributes were {', '.join(non_empty_log_attrs)}")
            raise AttributeError
    if is_stream:
        if is_cwltool_log:
            temp_html_obj = NamedTemporaryFile()
            write_websocket_to_file(
                log_stream,
                output_file=Path(temp_html_obj.name)
            )
            convert_html_to_text(Path(temp_html_obj.name), output_path)
        else:
            write_websocket_to_file(
                log_stream,
                output_file=output_path
            )
    else:
        read_icav2_file_contents(project_id, log_data_id, output_path)



#!/usr/bin/env python

# Import everything
from .functions.project_data_functions import (
    # Project Data functions
    get_project_data_file_id_from_project_id_and_path,
    create_data_in_project,
    create_file_in_project,
    create_folder_in_project,
    get_project_data_folder_id_from_project_id_and_path,
    get_project_data_id_from_project_id_and_path,
    get_project_data_obj_by_id,
    get_project_data_obj_from_project_id_and_path,
    get_project_data_path_by_id,
    list_project_data_non_recursively,
    find_project_data_recursively,
    find_project_data_bulk,
    create_download_url,
    create_download_urls,
    convert_icav2_uri_to_data_obj,
    convert_project_data_obj_to_icav2_uri,
    convert_project_id_and_data_path_to_icav2_uri,
    get_aws_credentials_access_for_project_folder,
    is_folder_id_format,
    is_file_id_format,
    is_data_id,
    presign_cwl_directory,
    presign_cwl_directory_with_external_data_mounts,
    read_icav2_file_contents,
    read_icav2_file_contents_to_string,
    get_project_data_upload_url,
    write_icav2_file_contents,
    get_file_by_file_name_from_project_data_list,
    project_data_copy_batch_handler
)

__all__ = [
    # Functions
    'get_project_data_file_id_from_project_id_and_path',
    'create_data_in_project',
    'create_file_in_project',
    'create_folder_in_project',
    'get_project_data_folder_id_from_project_id_and_path',
    'get_project_data_id_from_project_id_and_path',
    'get_project_data_obj_by_id',
    'get_project_data_obj_from_project_id_and_path',
    'get_project_data_path_by_id',
    'list_project_data_non_recursively',
    'find_project_data_recursively',
    'find_project_data_bulk',
    'create_download_url',
    'create_download_urls',
    'convert_icav2_uri_to_data_obj',
    'convert_project_data_obj_to_icav2_uri',
    'convert_project_id_and_data_path_to_icav2_uri',
    'get_aws_credentials_access_for_project_folder',
    'is_folder_id_format',
    'is_file_id_format',
    'is_data_id',
    'presign_cwl_directory',
    'presign_cwl_directory_with_external_data_mounts',
    'read_icav2_file_contents',
    'read_icav2_file_contents_to_string',
    'get_project_data_upload_url',
    'write_icav2_file_contents',
    'get_file_by_file_name_from_project_data_list',
    'project_data_copy_batch_handler'
]



#!/usr/bin/env python3


from libica.openapi.v2 import ApiClient, ApiException
from libica.openapi.v2.api.job_api import JobApi
from libica.openapi.v2.model.job import Job
from ...utils.configuration import get_icav2_configuration


def get_job(job_id: str) -> Job:
    configuration = get_icav2_configuration()

    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = JobApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve a job.
        api_response: Job = api_instance.get_job(job_id)
    except ApiException as e:
        raise ApiException("Exception when calling JobApi->get_job: %s\n" % e)

    return api_response



#!/usr/bin/env python3

"""
All the enums!
"""

from enum import Enum


class DataType(Enum):
    FOLDER = "FOLDER"
    FILE = "FILE"


class AnalysisStorageSize(Enum):
    SMALL = "SMALL"
    MEDIUM = "MEDIUM"
    LARGE = "LARGE"


class ProjectDataSortParameters(Enum):
    TIMECREATED = "timeCreated"
    TIMECREATED_DESC = "-timeCreated"
    TIMEMODIFIED = "timeModified"
    TIMEMODIFIED_DESC = "-timeModified"
    NAME = "name"
    NAME_DESC = "-name"
    PATH = "path"
    PATH_DESC = "-path"
    FILESIZEINBYTES = "fileSizeInBytes"
    FILESIZEINBYTES_DESC = "-fileSizeInBytes"
    STATUS = "status"
    STATUS_DESC = "-status"
    FORMAT = "format"
    FORMAT_DESC = "-format"
    DATATYPE = "dataType"
    DATATYPE_DESC = "-dataType"
    WILLBEARCHIVEDAT = "willBeArchivedAt"
    WILLBEARCHIVEDAT_DESC = "-willBeArchivedAt"
    WILLBEDELETEDAT = "willBeDeletedAt"
    WILLBEDELETEDAT_DESC = "-willBeDeletedAt"


class ProjectDataStatusValues(Enum):
    PARTIAL = "PARTIAL"
    AVAILABLE = "AVAILABLE"
    ARCHIVING = "ARCHIVING"
    ARCHIVED = "ARCHIVED"
    UNARCHIVING = "UNARCHIVING"
    DELETING = "DELETING"


class WorkflowLanguage(Enum):
    CWL = "CWL"
    NEXTFLOW = "NEXTFLOW"


class StructuredInputParameterType(Enum):
    BOOLEAN = "boolean"
    STRING = "string"
    INTEGER = "integer"
    OPTION = "options"


class StructuredInputParameterTypeMapping(Enum):
    BOOLEAN = bool
    STRING = str
    INTEGER = int
    OPTION = str



#!/usr/bin/env python3

"""
Generate a CWL analysis
"""
# Imports
import json
from pathlib import Path
from typing import List, Dict, Optional

# Libica imports
from libica.openapi.v2.model.analysis import Analysis
from libica.openapi.v2.model.analysis_input_data_mount import AnalysisInputDataMount
from libica.openapi.v2.model.analysis_input_external_data import AnalysisInputExternalData
from libica.openapi.v2.model.analysis_output_mapping import AnalysisOutputMapping
from libica.openapi.v2.model.create_cwl_analysis import CreateCwlAnalysis
from libica.openapi.v2.model.cwl_analysis_input import CwlAnalysisInput
from libica.openapi.v2.model.cwl_analysis_json_input import CwlAnalysisJsonInput

# Local parent imports
from .analysis import (
    ICAv2PipelineAnalysisTags,
    ICAv2PipelineAnalysis,
    ICAv2EngineParameters,
    ICAv2AnalysisInput
)

# Local imports
from ...enums import AnalysisStorageSize, WorkflowLanguage
from ...utils.globals import ICAv2AnalysisStorageSize
from ...utils.logger import get_logger

# Set logger
logger = get_logger()


class ICAv2CwlAnalysisJsonInput(ICAv2AnalysisInput):
    """
    Generate a CWL Analysis input from a JSON input
    """

    object_type = "JSON"

    def __init__(
        self,
        input_json: Dict
    ):
        # Initialise parent class
        super().__init__(
            object_type=self.object_type
        )

        # Set inputs
        self.input_json: Dict = input_json

        # Initialise values after dereferencing
        self.data_ids: List[str] = []
        self.input_json_deferenced_dict: Dict = {}
        self.mount_paths_list: List[AnalysisInputDataMount] = []
        self.external_mounts_list: List[AnalysisInputExternalData] = []

        # Generate typed values for creating analysis
        self.input_json_str: Optional[str] = None

    def validate_input(self):
        """
        Cannot yet validate the input as we don't have an easy way to collect a JSON schema for a CWL input
        :return:
        """
        pass

    def create_analysis_input(self) -> CwlAnalysisJsonInput:
        # Deference cwl input json
        self.deference_cwl_input_json()

        # Set input json to string
        self.set_input_json()

        # Generate a CWL analysis input
        return CwlAnalysisJsonInput(
            object_type=self.object_type,
            input_json=self.input_json_str,
            mounts=self.mount_paths_list,
            external_data=self.external_mounts_list
        )

    def deference_cwl_input_json(self):
        from ..functions.project_pipelines_functions import (
            convert_icav2_uris_to_data_ids_from_cwl_input_json
        )
        self.input_json_deferenced_dict, self.mount_paths_list, self.external_mounts_list = convert_icav2_uris_to_data_ids_from_cwl_input_json(
              self.input_json
        )

    def set_input_json(self):
        self.input_json_str = json.dumps(self.input_json_deferenced_dict, indent=2)


class ICAv2CWLEngineParameters(ICAv2EngineParameters):
    """
    The ICAv2 EngineParameters has the following properties
    """

    workflow_language = WorkflowLanguage.CWL

    def __init__(
        self,
        project_id: Optional[str] = None,
        pipeline_id: Optional[str] = None,
        output_parent_folder_id: Optional[str] = None,
        output_parent_folder_path: Optional[Path] = None,
        analysis_output: Optional[List[AnalysisOutputMapping]] = None,
        analysis_input: Optional[CwlAnalysisInput] = None,
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
        analysis_storage_id: Optional[str] = None,
        analysis_storage_size: Optional[ICAv2AnalysisStorageSize] = None,
        activation_id: Optional[str] = None,
        cwltool_overrides: Optional[Dict] = None
    ):
        # Initialise parameters
        super().__init__(
            project_id=project_id,
            pipeline_id=pipeline_id,
            output_parent_folder_id=output_parent_folder_id,
            output_parent_folder_path=output_parent_folder_path,
            analysis_output=analysis_output,
            analysis_input=analysis_input,
            tags=tags,
            analysis_storage_id=analysis_storage_id,
            analysis_storage_size=analysis_storage_size,
            activation_id=activation_id
        )

        # Input parameters
        self.cwltool_overrides: Dict = cwltool_overrides
        # self.stream_all_files: Optional[bool] = stream_all_files
        # self.stream_all_directories: Optional[bool] = stream_all_directories

    def get_overrides_from_engine_parameters(self, inputs_overrides: Optional[Dict] = None) -> Dict:
        """
        Get the overrides from the engine parameters.
        Note overrides in the inputjson should NOT be overwritten
        :param inputs_overrides:
        :return:
        """

        # If overrides are in the engine parameters, put them in the input json
        if inputs_overrides is None:
            inputs_overrides = {}

        elif inputs_overrides is not None and not isinstance(inputs_overrides, Dict):
            logger.error(f"inputs_overrides was not a dict, instead type {type(inputs_overrides)}")
            raise TypeError("inputs_overrides was not a dict")

        # Initialise the engine parameters cwltool overrides dict
        engine_parameter_cwltool_overrides = self.cwltool_overrides.copy()

        # If there are no engine parameter overrides, return the input json overrides
        if engine_parameter_cwltool_overrides is None or len(engine_parameter_cwltool_overrides) == 0:
            return inputs_overrides

        # Don't override existing overrides in the input json
        # Instead overwrite engine parameter cwltool overrides
        for key, value in inputs_overrides.items():
            if key in engine_parameter_cwltool_overrides.keys():
                engine_parameter_cwltool_overrides[key].update(value)
            else:
                engine_parameter_cwltool_overrides[key] = value

        # Then write them back in here again
        # Now we pull them back in again
        for key, value in engine_parameter_cwltool_overrides.items():
            inputs_overrides[key] = value

        return inputs_overrides


class ICAv2CWLPipelineAnalysis(ICAv2PipelineAnalysis):
    """
    The ICAv2CWLPipelineAnalysis has the following properties
        * user_reference: str
        * input_json: Dict  (cwl_inputs)
        * engine_parameters: ICAv2EngineParameters

    The engineParameters then populate the following parameters for its parent class

    pipeline_id: str,
    tags: ICAv2PipelineAnalysisTags,
    analysis_input: Union[CwlAnalysisJsonInput, NextflowAnalysisInput],
    analysis_output: str,
    activation_code_details_id: Optional[str] = None,
    analysis_storage_size: Optional[AnalysisStorageSize] = None

    """

    def __init__(
        self,
        # Launch parameters
        user_reference: str,
        project_id: str,
        pipeline_id: str,
        analysis_input: CwlAnalysisInput,
        analysis_storage_id: Optional[str] = None,
        analysis_storage_size: Optional[AnalysisStorageSize] = None,
        activation_id: Optional[str] = None,
        # Output parameters
        output_parent_folder_id: Optional[str] = None,
        output_parent_folder_path: Optional[str] = None,
        analysis_output_uri: Optional[str] = None,
        # Meta parameters
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
        # CWL Specific parameters
        cwltool_overrides: Optional[Dict] = None
    ):
        """
        Initialise input
        :param user_reference
        :param project_id
        :param pipeline_id
        :param analysis_input
        :param analysis_storage_id
        :param analysis_storage_size
        :param activation_id
        :param output_parent_folder_id
        :param output_parent_folder_path
        :param analysis_output_uri
        :param tags
        :param cwltool_overrides
        """
        # Initialise any cwl specific parameters first before calling the parent class
        # Set cwl specific inputs
        self.cwltool_overrides: Optional[Dict] = cwltool_overrides
        # Set under parent init script through set_engine_parameters
        self.engine_parameters: Optional[ICAv2CWLEngineParameters] = None

        # Call parent class
        super().__init__(
            user_reference=user_reference,
            project_id=project_id,
            pipeline_id=pipeline_id,
            analysis_input=analysis_input,
            analysis_storage_id=analysis_storage_id,
            analysis_storage_size=analysis_storage_size,
            activation_id=activation_id,
            output_parent_folder_id=output_parent_folder_id,
            output_parent_folder_path=output_parent_folder_path,
            analysis_output_uri=analysis_output_uri,
            tags=tags
        )

    def set_engine_parameters(self):
        self.engine_parameters = ICAv2CWLEngineParameters(
            project_id=self.project_id,
            pipeline_id=self.pipeline_id,
            output_parent_folder_id=self.output_parent_folder_id,
            output_parent_folder_path=self.output_parent_folder_path,
            analysis_output=self.analysis_output,
            analysis_input=self.analysis_input,
            tags=self.tags,
            analysis_storage_id=self.analysis_storage_id,
            analysis_storage_size=self.analysis_storage_size,
            activation_id=self.activation_id,
            cwltool_overrides=self.cwltool_overrides
        )

    def create_analysis(self) -> CreateCwlAnalysis:
        return CreateCwlAnalysis(
            user_reference=self.user_reference,
            pipeline_id=self.pipeline_id,
            tags=self.engine_parameters.tags(),
            activation_code_detail_id=self.engine_parameters.activation_id,
            analysis_input=self.analysis_input,
            analysis_storage_id=self.engine_parameters.analysis_storage_id,
            output_parent_folder_id=self.engine_parameters.output_parent_folder_id,
            analysis_output=self.engine_parameters.analysis_output
        )

    def launch_analysis(self) -> Analysis:
        from ..functions.project_pipelines_functions import launch_cwl_workflow
        return launch_cwl_workflow(
            project_id=self.project_id,
            cwl_analysis=self.analysis,
        )



#!/usr/bin/env python3

"""
Intialise the configuration for the application
"""
# Standard imports
from datetime import datetime
from pathlib import Path
from os import environ
from typing import Optional, OrderedDict
from urllib.parse import urlparse
from jwt import decode, InvalidTokenError
from ruamel.yaml import YAML

# Libica imports
from libica.openapi.v2 import Configuration

# Local imports
from .globals import ICAV2_CONFIG_FILE_PATH, ICAV2_CONFIG_FILE_SERVER_URL_KEY, DEFAULT_ICAV2_BASE_URL, \
    ICAV2_ACCESS_TOKEN_AUDIENCE, ICAV2_SESSION_FILE_PATH, ICAV2_SESSION_FILE_ACCESS_TOKEN_KEY
from .logger import get_logger
from .subprocess_handler import run_subprocess_proc

# Set logger
logger = get_logger()

# Global runtime vars
ICAV2_CONFIGURATION: Optional[Configuration] = None


# Read the configuration file
def get_config_file_path() -> Path:
    """
    Get path for the config file and asser it exists
    :return:
    """
    config_file_path: Path = Path(ICAV2_CONFIG_FILE_PATH.format(
      HOME=environ["HOME"]
    ))

    if not config_file_path.is_file():
        logger.error(f"Could not get file path {config_file_path}")
        raise FileNotFoundError

    return config_file_path


def read_config_file() -> OrderedDict:
    """
    Get the contents of the session file (~/.icav2/config.ica.yaml)
    :return:
    """

    logger.debug("Reading in the config file")
    yaml = YAML()

    with open(get_config_file_path(), "r") as file_h:
        data = yaml.load(file_h)

    return data


# Set the icav2 environment variables
def get_icav2_base_url():
    """
    Collect the icav2 base url for the configuration
    Likely 'https://ica.illumina.com/ica/rest'
    :return:
    """
    # Check env
    icav2_base_url_env = environ.get("ICAV2_BASE_URL", None)
    if icav2_base_url_env is not None:
        return icav2_base_url_env

    # Read config file
    config_yaml_dict = read_config_file()
    if ICAV2_CONFIG_FILE_SERVER_URL_KEY in config_yaml_dict.keys():
        return f"https://{config_yaml_dict[ICAV2_CONFIG_FILE_SERVER_URL_KEY]}/ica/rest"
    else:
        logger.warning("Could not get server-url from config yaml")

    return DEFAULT_ICAV2_BASE_URL


def refresh_access_token() -> str:
    """
    Run standard command to get a new token in the session file
    :return:
    """
    project_list_returncode, project_list_stdout, project_list_stderr = \
        run_subprocess_proc(["icav2", "projects", "list"], capture_output=True)

    if not project_list_returncode == 0:
        logger.error("Couldn't run a simple icav2 command to refresh the token")
        raise ChildProcessError

    # Get the newly refreshed token
    return get_access_token_from_session_file(refresh=False)


def check_access_token_expiry(access_token: str) -> bool:
    """
    Check access token hasn't expired
    True if has not expired, otherwise false
    :param access_token:
    :return:
    """
    current_epoch_time = int(datetime.now().strftime("%s"))

    if current_epoch_time < get_jwt_token_obj(access_token, ICAV2_ACCESS_TOKEN_AUDIENCE).get("exp"):
        return True

    # Otherwise
    logger.info("Token has expired")
    return False


def get_session_file_path() -> Path:
    """
    Get path for session file and assert file exists
    :return:
    """
    session_file_path: Path = Path(ICAV2_SESSION_FILE_PATH.format(
        HOME=environ["HOME"],
        server_url_prefix=urlparse(get_icav2_base_url()).netloc.split(".")[0]
    ))

    if not session_file_path.is_file():
        logger.error(f"Could not get file path {session_file_path}")
        raise FileNotFoundError

    return session_file_path


def read_session_file() -> OrderedDict:
    """
    Get the contents of the session file (~/.icav2/.session.ica.yaml)
    :return:
    """

    logger.debug("Reading in the session file")
    yaml = YAML()

    with open(get_session_file_path(), "r") as file_h:
        data = yaml.load(file_h)

    return data


def get_access_token_from_session_file(refresh: bool = True) -> str:
    """
    Collect the contents of the access token
    :return:
    """
    session_data: OrderedDict = read_session_file()

    access_token: str = session_data.get(ICAV2_SESSION_FILE_ACCESS_TOKEN_KEY, None)

    if access_token is None:
        logger.error("Could not get access token from session file")
        raise KeyError

    if not check_access_token_expiry(access_token):
        if not refresh:
            logger.error(f"Could not refresh access token in session file {ICAV2_SESSION_FILE_PATH}")
            raise KeyError
        else:
            access_token = refresh_access_token()

    return access_token


def get_icav2_access_token() -> str:
    """
    Return icav2 access token
    """
    access_token = environ.get("ICAV2_ACCESS_TOKEN", None)

    if access_token is None or not check_access_token_expiry(access_token):
        access_token = get_access_token_from_session_file()

    return access_token


def set_icav2_configuration():
    global ICAV2_CONFIGURATION

    ICAV2_CONFIGURATION = Configuration(
        host=get_icav2_base_url(),
        access_token=get_icav2_access_token()
    )


def get_icav2_configuration() -> Configuration:
    """
    Return icav2 configuration, if not set, sets it first, then returns
    :return:
    """
    if ICAV2_CONFIGURATION is None:
        set_icav2_configuration()
    return ICAV2_CONFIGURATION


def get_jwt_token_obj(jwt_token, audience):
    """
    Get the jwt token object through the pyjwt package
    :param jwt_token: The jwt token in base64url format
    :param audience: The Audience to use for the token, defaults to 'ica'
    :return:
    """
    try:
        token_object = decode(jwt_token,
                              options={
                                         "verify_signature": False,
                                         "require_exp": True
                                       },
                              audience=[audience],
                              algorithms="RS256",
                              )
    except InvalidTokenError:
        raise InvalidTokenError

    return token_object




#!/usr/bin/env python

"""
An agnostic Analysis Object

Both CWL Analysis and NextFlow analysis will use this object
as their parent class

"""
# Standard imports
import json
from pathlib import Path
from typing import List, Optional, Union, Dict, Any
from urllib.parse import urlparse
from datetime import datetime
from uuid import uuid4

# Libica imports
from libica.openapi.v2.model.analysis import Analysis
from libica.openapi.v2.model.analysis_output_mapping import AnalysisOutputMapping
from libica.openapi.v2.model.analysis_tag import AnalysisTag
from libica.openapi.v2.model.create_cwl_analysis import CreateCwlAnalysis
from libica.openapi.v2.model.create_nextflow_analysis import CreateNextflowAnalysis
from libica.openapi.v2.model.cwl_analysis_input import CwlAnalysisInput
from libica.openapi.v2.model.cwl_analysis_json_input import CwlAnalysisJsonInput
from libica.openapi.v2.model.cwl_analysis_structured_input import CwlAnalysisStructuredInput
from libica.openapi.v2.model.nextflow_analysis_input import NextflowAnalysisInput
from libica.openapi.v2.model.project_data import ProjectData

# Local imports
from ...utils import recursively_build_open_api_body_from_libica_item
from ...enums import AnalysisStorageSize, DataType
from ...utils.globals import ICAv2AnalysisStorageSize
from ...utils.miscell import sanitise_dict_keys
from ...utils.logger import get_logger

# Set logger
logger = get_logger()


class ICAv2AnalysisInput:
    """
    Parent class for CWLAnalysisInput, NextflowAnalysisInput
    """
    def __init__(
        self,
        object_type: str,
    ):
        # Set the object type
        self.object_type = object_type

    def __call__(self) -> Union[CwlAnalysisInput, CwlAnalysisJsonInput, CwlAnalysisStructuredInput, NextflowAnalysisInput]:
        """
        Create the analysis input
        :return:
        """
        self.validate_input()
        return self.create_analysis_input()

    def create_analysis_input(self):
        """
        Implemented in subclass
        :return:
        """
        raise NotImplementedError

    def validate_input(self):
        """
        For structured inputs, one can first import the pipeline id and compare the input values against those
        specified in the pipeline, and compare required against non-required inputs
        For JSON based inputs, one will need to validate against the JSON schema. This is not implemented yet
        as waiting on PR https://github.com/common-workflow-language/cwl-utils/pull/288
        We also want https://github.com/umccr-illumina/ica_v2/issues/162 resolved so we can validate against the top level workflow object.
        :return:
        """
        raise NotImplementedError


class ICAv2PipelineAnalysisTags:
    """
    List of tags
    """
    def __init__(
        self,
        technical_tags: Union[List | Dict],
        user_tags: Union[List | Dict],
        reference_tags: Union[List | Dict]
    ):
        """
        List of tags to use in the pipeline
        :param technical_tags:
        :param user_tags:
        :param reference_tags:
        """

        # Assign
        self.technical_tags = technical_tags
        self.user_tags = user_tags
        self.reference_tags = reference_tags

        self.clean_tags()

    def clean_tags(self):
        # Now clean up
        for tag_type in ["technical_tags", "user_tags", "reference_tags"]:
            if not isinstance(getattr(self, tag_type), (list, dict)):
                raise ValueError(f"{tag_type} must be a list or a dictionary")
            # For each key-value pair in the dictionary, convert it to a string split by '='
            if isinstance(getattr(self, tag_type), dict):
                setattr(
                    self,
                    tag_type,
                    list(
                        map(
                            lambda tag_iter: f"{tag_iter[0]}={tag_iter[1]}",
                            getattr(self, tag_type).items()
                        )
                    )
                )
            else:
                # Ensure that each item in the tag list is a string
                setattr(
                    self,
                    tag_type,
                    list(
                        map(
                            str,
                            getattr(self, tag_type)
                        )
                    )
                )

    def __call__(self) -> AnalysisTag:
        return AnalysisTag(
            technical_tags=self.technical_tags,
            user_tags=self.user_tags,
            reference_tags=self.reference_tags
        )

    def combine_tags(self, analysis_tags: 'ICAv2PipelineAnalysisTags') -> 'ICAv2PipelineAnalysisTags':
        """
        Combine tags from another tag object
        :param analysis_tags:
        :return:
        """
        # Combine the tags
        return ICAv2PipelineAnalysisTags(
            technical_tags=self.technical_tags + analysis_tags.technical_tags,
            user_tags=self.user_tags + analysis_tags.user_tags,
            reference_tags=self.reference_tags + analysis_tags.reference_tags
        )

    @classmethod
    def from_dict(cls, tags_dict):
        # Convert camel cases to snake cases
        tags_dict = sanitise_dict_keys(tags_dict)

        return cls(
                    technical_tags=tags_dict.get("technical_tags", []),
                    user_tags=tags_dict.get("user_tags", []),
                    reference_tags=tags_dict.get("reference_tags", [])
        )


class ICAv2EngineParameters:
    """
    The ICAv2 EngineParameters has the following properties
    """

    # Set in subclasses
    workflow_language = None

    def __init__(
            self,
            project_id: Optional[str] = None,
            pipeline_id: Optional[str] = None,
            output_parent_folder_id: Optional[str] = None,
            output_parent_folder_path: Optional[Path] = None,
            analysis_input: Optional[Union[CwlAnalysisInput, NextflowAnalysisInput]] = None,
            analysis_output: Optional[List[AnalysisOutputMapping]] = None,
            tags: Optional[ICAv2PipelineAnalysisTags] = None,
            analysis_storage_id: Optional[str] = None,
            analysis_storage_size: Optional[ICAv2AnalysisStorageSize] = None,
            activation_id: Optional[str] = None,
    ):
        # Initialise parameters

        # Launch parameters
        self.project_id: Optional[str] = project_id
        self.pipeline_id: Optional[str] = pipeline_id
        self.analysis_storage_id: Optional[str] = analysis_storage_id
        self.analysis_storage_size: Optional[ICAv2AnalysisStorageSize] = analysis_storage_size
        self.activation_id: Optional[str] = activation_id

        # Output parameters
        self.output_parent_folder_id: Optional[str] = output_parent_folder_id
        self.output_parent_folder_path: Optional[Path] = output_parent_folder_path
        self.analysis_output: Optional[List[AnalysisOutputMapping]] = analysis_output
        self.analysis_input: Optional[Union[CwlAnalysisInput, NextflowAnalysisInput]] = analysis_input

        # Meta parameters
        self.tags: ICAv2PipelineAnalysisTags = tags

        # self.stream_all_files: Optional[bool] = stream_all_files
        # self.stream_all_directories: Optional[bool] = stream_all_directories

        # Set placeholders
        current_utc_time = datetime.utcnow()
        self.placeholder_dict: Dict = {
            "__DATE_STR__": current_utc_time.strftime("%Y%m%d"),
            "__TIME_STR__": current_utc_time.strftime("%H%M%S"),
            "__UUID8_STR__": uuid4().hex[:8],
            "__UUID16_STR__": uuid4().hex[:16],
        }

    def __call__(self):
        # Assumed that the following have been run
        # set_launch_parameters
        # set_output_parameters
        # set_meta_parameters
        # All functions can be harmlessly called again
        self.set_launch_parameters()
        self.set_output_parameters()
        self.set_meta_parameters()

        # Now check parameters
        self.check_launch_parameters()
        self.check_output_parameters()
        self.check_meta_parameters()

    def set_launch_parameters(
        self,
        project_id: Optional[str] = None,
        pipeline_id: Optional[str] = None,
        analysis_storage_id: Optional[str] = None,
        analysis_storage_size: Optional[ICAv2AnalysisStorageSize] = None,
        activation_id: Optional[str] = None,
        analysis_input: Optional[Union[CwlAnalysisJsonInput, CwlAnalysisStructuredInput]] = None
    ):
        # Local import of functions from classes to avoid circular imports
        from ..functions.project_pipelines_functions import (
            get_default_analysis_storage_id_from_project_pipeline,
            get_activation_id,
            get_analysis_storage_id_from_analysis_storage_size
        )
        # Check for project id
        if project_id is not None:
            self.project_id = project_id

        # Check for pipeline id
        if pipeline_id is not None:
            self.pipeline_id = pipeline_id

        # Check for analysis storage id
        if analysis_storage_id is not None:
            self.analysis_storage_id = analysis_storage_id

        # Check for analysis storage size
        if analysis_storage_size is not None:
            self.analysis_storage_size = analysis_storage_size

        # Check for analysis input
        if analysis_input is not None:
            self.analysis_input = analysis_input

        # If analysis storage size is set, but analysis storage id is not set
        if self.analysis_storage_size is not None and self.analysis_storage_id is None:
            self.analysis_storage_id = get_analysis_storage_id_from_analysis_storage_size(
                self.analysis_storage_size
            )

        elif self.analysis_storage_size is None and self.analysis_storage_id is None:
            # Ensure that project_id and pipeline_id are set at this point
            if self.project_id is None or self.pipeline_id is None:
                logger.error("Cannot get analysis storage id without project id and pipeline id")
                raise ValueError
            # Get the default analysis storage id
            self.analysis_storage_id = get_default_analysis_storage_id_from_project_pipeline(
                self.project_id,
                self.pipeline_id
            )

        # Check for activation id
        if activation_id is not None:
            self.activation_id = activation_id
        elif self.activation_id is None:
            # Ensure that both project_id and pipeline_id are set at this point
            if self.project_id is None or self.pipeline_id is None or self.analysis_input is None:
                logger.error("Cannot get the activation id without project id and pipeline id or analysis input")
                raise ValueError
            self.activation_id = get_activation_id(
                project_id=self.project_id,
                pipeline_id=self.pipeline_id,
                analysis_input=self.analysis_input,
                workflow_language=self.workflow_language
            )

    def check_launch_parameters(self):
        """
        Check the launch parameters have been set
        :return:
        """
        has_error = False
        for parameter in ["project_id", "pipeline_id", "analysis_storage_id", "activation_id"]:
            if self.__getattribute__(parameter) is None:
                logger.error(f"{parameter} has not been set")
                has_error = True
        if has_error:
            raise ValueError

    def set_output_parameters(
        self,
        output_parent_folder_id: Optional[str] = None,
        output_parent_folder_path: Optional[Path] = None,
        analysis_output: Optional[List[AnalysisOutputMapping]] = None
    ):
        """
        Select one of output parent folder id, output parent folder path or analysis output
        :param output_parent_folder_id:
        :param output_parent_folder_path:
        :param analysis_output:
        :return:
        """
        from ...project_data import get_project_data_folder_id_from_project_id_and_path

        # Just set the id if it is not None
        if output_parent_folder_id is not None:
            self.output_parent_folder_id = output_parent_folder_id
        # Update output parent folder id with output parent folder path
        # Need to check if project_id has been set
        elif (
                self.output_parent_folder_id is None and
                (
                    self.output_parent_folder_path is not None
                    or output_parent_folder_path is not None
                )
        ):
            # Need to check if project_id has been set
            if self.project_id is None:
                logger.error(
                    "project_id has not been set, cannot set output_parent_folder_id from folder path "
                    "without project_id"
                )
                raise ValueError

            # Override existing with output in here
            if output_parent_folder_path is not None:
                self.output_parent_folder_path = output_parent_folder_path

            # Populate placeholders in the output path
            self.output_parent_folder_path = self.populate_placeholders_in_output_path(self.output_parent_folder_path)

            self.output_parent_folder_id = get_project_data_folder_id_from_project_id_and_path(
                project_id=self.project_id,
                folder_path=self.output_parent_folder_path,
                create_folder_if_not_found=True
            )

        # Use the analysis output instead
        elif analysis_output is not None:
            self.analysis_output = analysis_output

    def check_output_parameters(self):
        """
        Ensure that at least one of output_parent_folder_id and analysis_output is set
        :return:
        """
        # Ensure that at least one of output_parent_folder_id and analysis_output is set
        if (
            self.output_parent_folder_id is None and self.analysis_output is None
        ):
            # Needed to specify one of the inputs
            logger.error("Not one of output_parent_folder_id and analysis_output is set")
            raise ValueError

    def set_meta_parameters(
        self,
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
    ):
        # Set tags if tags is empty
        if tags is not None and self.tags is None:
            self.tags = tags

        # Combine tags if both are set
        elif tags is not None and self.tags is not None:
            self.tags = self.tags.combine_tags(tags)

    def check_meta_parameters(self):
        if self.tags is None:
            logger.warning("No tags set for this analysis")

    def update_engine_parameter(self, attribute_name: str, value: Any):
        self.__setattr__(attribute_name, value)

    def populate_placeholders_in_output_path(self, analysis_path: Path):
        """
        Populate placeholders in the output path
        :param analysis_path:
        :return:
        """
        # Import functions locally to avoid circular imports
        from ...utils import fill_placeholder_path
        return fill_placeholder_path(
            analysis_path,
            self.placeholder_dict
        )


class ICAv2PipelineAnalysis:
    """
    A parent analysis class with the following attributes

    * name / user-reference  # The pipeline name
    * pipeline-id  # The pipeline id
    * tags - Tags used on the analysis ICAv2PipelineAnalysisTags class
    * activationCodeDetailsId - Collection of the activation code is implemented in the subclass
    * analysisStorageId
    * analysisOutput ( we skip the outputParentFolderId, it doesn't give us much control)
        * This should be a path-based uri
    * analysisInput ( implemented in the subclass )
    """

    def __init__(
        self,
        # Launch parameters
        user_reference: str,
        project_id: str,
        pipeline_id: str,
        analysis_input: Union[CwlAnalysisInput, NextflowAnalysisInput],
        analysis_storage_id: Optional[str] = None,
        analysis_storage_size: Optional[AnalysisStorageSize] = None,
        activation_id: Optional[str] = None,
        # Output parameters
        output_parent_folder_id: Optional[str] = None,
        output_parent_folder_path: Optional[str] = None,
        analysis_output_uri: Optional[str] = None,
        # Meta parameters
        tags: Optional[ICAv2PipelineAnalysisTags] = None
    ):
        """
        Initialise parent analysis class with the following attributes

        * name / user-reference  # The analysis name - Required
        * project-id  # The project id - Required
        * pipeline-id  # The pipeline id - Required
        * tags - Tags used on the analysis ICAv2PipelineAnalysisTags class  - Required
        * analysis_storage_size - The size of the analysis storage - Optional
            * We can figure this out later
        * activation_id - Collection of the activation code is implemented in the subclass - Optional
            * We can figure this out later
        * analysisOutput ( we skip the outputParentFolderId, it doesn't give us much control)
            * This should be a path-based uri - Required
        * analysis_input ( implemented in the subclass )
        """
        # Launch parameters
        self.user_reference = user_reference
        self.project_id = project_id
        self.pipeline_id = pipeline_id
        self.activation_id = activation_id
        self.output_parent_folder_id = output_parent_folder_id
        self.output_parent_folder_path = output_parent_folder_path
        self.analysis_output_uri = analysis_output_uri

        # Meta parameters
        self.tags = tags
        self.analysis_storage_id = analysis_storage_id
        self.analysis_storage_size = analysis_storage_size
        self.analysis_input = analysis_input

        self.engine_parameters: Optional[ICAv2EngineParameters] = None

        if self.analysis_output_uri is not None:
            self.analysis_output: List[AnalysisOutputMapping] = [self.get_analysis_output_mapping_from_uri()]
        else:
            self.analysis_output = None

        self.set_engine_parameters()

        # Set the analysis
        self.analysis: Optional[Union[CreateCwlAnalysis, CreateNextflowAnalysis]] = None

    def __call__(self) -> Analysis:
        """
        Fix up the engine parameters, and generate a CWL Analysis Object
        :return:
        """
        # Check engine parameters
        self.check_engine_parameters()

        # Create analysis
        self.analysis = self.create_analysis()

        return self.launch_analysis()

    def get_analysis_output_mapping_from_uri(self) -> AnalysisOutputMapping:
        """
        Convert the analysis output to a mapping
        :return:
        """
        from ...project_data import convert_icav2_uri_to_data_obj
        # Ensure that the path attribute of analysis_output_uri ends with /
        if not urlparse(self.analysis_output_uri).path.endswith("/"):
            raise ValueError("The analysis output uri must end with a /")

        analysis_output_obj: ProjectData = convert_icav2_uri_to_data_obj(
            self.analysis_output_uri,
            create_data_if_not_found=True
        )

        return AnalysisOutputMapping(
            source_path="out/",  # Hardcoded, all workflow outputs should be placed in the out folder,
            type=DataType.FOLDER.value,  # Hardcoded, out directory is a folder
            target_project_id=analysis_output_obj.project_id,
            target_path=analysis_output_obj.data.details.path
        )

    def set_engine_parameters(self):
        """
        Implemented in subclass
        :return:
        """
        raise NotImplementedError

    def check_engine_parameters(self):
        """
        Implemented in subclass
        :return:
        """
        self.engine_parameters()

    def create_analysis(self) -> Union[CreateCwlAnalysis, CreateNextflowAnalysis]:
        # Implemented in subclass
        raise NotImplementedError

    def launch_analysis(self) -> Analysis:
        # Implemented in subclass
        raise NotImplementedError

    def get_analysis_launch_json(self) -> Dict:
        """
        Save the analysis launch json
        :return:
        """
        return recursively_build_open_api_body_from_libica_item(
            self.analysis
        )

    def save_analysis(self, json_path: Path):
        """
        Save the analysis to a json file
        :param json_path:
        :return:
        """
        analysis_dict = self.get_analysis_launch_json()

        with open(json_path, "w") as f:
            f.write(
                json.dumps(
                    analysis_dict,
                    indent=2
                )
            )



#!/usr/bin/env python3

"""
Nextflow analysis
"""

# Imports
from pathlib import Path
from typing import List, Dict, Optional

# Libica imports
from libica.openapi.v2.model.analysis import Analysis
from libica.openapi.v2.model.analysis_data_input import AnalysisDataInput
from libica.openapi.v2.model.analysis_output_mapping import AnalysisOutputMapping
from libica.openapi.v2.model.analysis_parameter_input import AnalysisParameterInput
from libica.openapi.v2.model.create_nextflow_analysis import CreateNextflowAnalysis
from libica.openapi.v2.model.input_parameter import InputParameter
from libica.openapi.v2.model.nextflow_analysis_input import NextflowAnalysisInput
from libica.openapi.v2.model.pipeline_configuration_parameter import PipelineConfigurationParameter

# Local parent imports
from .analysis import (
    ICAv2PipelineAnalysisTags,
    ICAv2PipelineAnalysis,
    ICAv2EngineParameters,
    ICAv2AnalysisInput
)

# Local imports
from ...enums import (
    AnalysisStorageSize, WorkflowLanguage,
    StructuredInputParameterType, StructuredInputParameterTypeMapping
)
from ...utils.globals import ICAv2AnalysisStorageSize
from ...utils.logger import get_logger

# Set logger
logger = get_logger()


class ICAv2NextflowAnalysisInput(ICAv2AnalysisInput):
    """
    Generate a Nextflow Analysis input from a JSON input - 
    if the input is an icav2 uri, it is considered an analysis data input
    """

    object_type = "STRUCTURED"

    def __init__(
        self,
        input_json: Dict,
        pipeline_id: str,
        project_id: str
    ):
        # Initialise parent class
        super().__init__(
            object_type=self.object_type
        )

        # Set inputs
        self.input_json: Dict = input_json

        # Insert pipeline id and project id
        self.pipeline_id: str = pipeline_id
        self.project_id: str = project_id

        # Split input types
        self.inputs: List[AnalysisDataInput] = []
        self.parameters: List[AnalysisParameterInput] = []

    def validate_input(self):
        """
        Validate the input parameters much their expected inputs from the pipeline
        :return:
        """
        from ..functions.project_pipelines_functions import (
            get_project_pipeline_input_parameters,
            get_project_pipeline_configuration_parameters
        )

        # Get the pipeline input parameters
        pipeline_input_parameters: List[InputParameter] = get_project_pipeline_input_parameters(
            project_id=self.project_id,
            pipeline_id=self.pipeline_id
        )

        # Get the pipeline configuration parameters
        pipeline_configuration_parameters: List[PipelineConfigurationParameter] = (
            get_project_pipeline_configuration_parameters(
                project_id=self.project_id,
                pipeline_id=self.pipeline_id
            )
        )

        # With the input parameters, we can validate against the input json
        errors_list = []
        for input_parameter in pipeline_input_parameters:
            # Check required inputs
            if input_parameter.required:
                if not any(
                    list(
                        map(
                            lambda input_iter: input_iter.parameter_code == input_parameter.code,
                            self.inputs
                        )
                    )
                ):
                    logger.error(f"Required input parameter {input_parameter.code} not found in input json")
                    errors_list.append(f"Required input parameter {input_parameter.code} not found in input json")

            # Check if any multi-values allow for multiple values
            if not input_parameter.multi_value:
                if (
                    len(
                        list(
                            filter(
                                lambda input_iter: input_iter.parameter_code == input_parameter.code,
                                self.inputs
                            )
                        )
                    ) > 1
                ):
                    logger.error(f"non-multiple input parameter {input_parameter.code} specified multiple times")
                    errors_list.append(f"non-multiple input parameter {input_parameter.code} specified multiple times")

        # Iterate through configuration parameters
        for configuration_parameter in pipeline_configuration_parameters:
            # Check if required parameter is in configuration list
            if configuration_parameter.required:
                if not any(
                    list(
                        map(
                            lambda parameter_iter: parameter_iter.code == configuration_parameter.code,
                            self.parameters
                        )
                    )
                ):
                    logger.error(f"Required configuration parameter {configuration_parameter.code} not found in input json")
                    errors_list.append(f"Required configuration parameter {configuration_parameter.code} not found in input json")

            # Check if any multi-values allow for multiple values
            if not configuration_parameter.multi_value:
                if (
                    len(
                        list(
                            filter(
                                lambda parameter_iter: parameter_iter.code == configuration_parameter.code,
                                self.parameters
                            )
                        )
                    ) > 1
                ):
                    logger.error(f"non-multiple configuration parameter {configuration_parameter.code} specified multiple times")
                    errors_list.append(f"non-multiple configuration parameter {configuration_parameter.code} specified multiple times")

            # Check type of parameter matches the type in the configuration
            for parameter in self.parameters:
                if parameter.code == configuration_parameter.code:
                    if not (
                        # Check if the parameter value is of the expected configuration parameter type
                        isinstance(
                            parameter.value,
                            StructuredInputParameterTypeMapping[
                                StructuredInputParameterType(
                                    configuration_parameter.type
                                ).name
                            ].value
                        )
                    ):
                        logger.error(f"Parameter {parameter.code} is not of type {configuration_parameter.type}")
                        errors_list.append(f"Parameter {parameter.code} is not of type {configuration_parameter.type}")

            if len(errors_list) == 0:
                logger.info("Inputs validation passed")
                return

            logger.error("Inputs validation failed")
            for error in errors_list:
                logger.error(error)
            raise ValueError("Inputs validation failed")

    def create_analysis_input(self) -> NextflowAnalysisInput:
        # Set input json to string
        self.split_input_json_by_inputs_and_parameters()

        self.validate_input()

        # Generate a CWL analysis input
        return NextflowAnalysisInput(
            inputs=self.inputs,
            parameters=self.parameters,
        )

    def split_input_json_by_inputs_and_parameters(self):
        # Local imports for functions
        from ...project_data import convert_icav2_uri_to_data_obj
        for key, value in self.input_json.items():

            if isinstance(value, List):
                if len(value) == 0:
                    continue
                if value[0].startswith("icav2://"):
                    self.inputs.append(
                        AnalysisDataInput(
                            parameter_code=key,
                            data_ids=list(
                                map(
                                    lambda icav2_uri_iter: convert_icav2_uri_to_data_obj(icav2_uri_iter).data.id,
                                    value
                                )
                            )
                        )
                    )
                else:
                    self.parameters.append(
                        AnalysisParameterInput(
                            code=key,
                            multi_value=value,
                        )
                    )
            else:
                if isinstance(value, str) and value.startswith("icav2://"):
                    self.inputs.append(
                        AnalysisDataInput(
                            parameter_code=key,
                            data_ids=[convert_icav2_uri_to_data_obj(value).data.id]
                        )
                    )
                else:
                    self.parameters.append(
                        AnalysisParameterInput(
                            code=key,
                            value=str(value)
                        )
                    )


class ICAv2NextflowEngineParameters(ICAv2EngineParameters):
    """
    The ICAv2 EngineParameters has the following properties
    """

    workflow_language = WorkflowLanguage.NEXTFLOW

    def __init__(
        self,
        project_id: Optional[str] = None,
        pipeline_id: Optional[str] = None,
        output_parent_folder_id: Optional[str] = None,
        output_parent_folder_path: Optional[Path] = None,
        analysis_output: Optional[List[AnalysisOutputMapping]] = None,
        analysis_input: Optional[NextflowAnalysisInput] = None,
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
        analysis_storage_id: Optional[str] = None,
        analysis_storage_size: Optional[ICAv2AnalysisStorageSize] = None,
        activation_id: Optional[str] = None,
    ):
        # Initialise parameters
        super().__init__(
            project_id=project_id,
            pipeline_id=pipeline_id,
            output_parent_folder_id=output_parent_folder_id,
            output_parent_folder_path=output_parent_folder_path,
            analysis_output=analysis_output,
            analysis_input=analysis_input,
            tags=tags,
            analysis_storage_id=analysis_storage_id,
            analysis_storage_size=analysis_storage_size,
            activation_id=activation_id
        )


class ICAv2NextflowPipelineAnalysis(ICAv2PipelineAnalysis):
    """
    The ICAv2CWLPipelineAnalysis has the following properties
        * user_reference: str
        * input_json: Dict  (cwl_inputs)
        * engine_parameters: ICAv2EngineParameters

    The engineParameters then populate the following parameters for its parent class

    pipeline_id: str,
    tags: ICAv2PipelineAnalysisTags,
    analysis_input: Union[CwlAnalysisJsonInput, NextflowAnalysisInput],
    analysis_output: str,
    activation_code_details_id: Optional[str] = None,
    analysis_storage_size: Optional[AnalysisStorageSize] = None

    """

    def __init__(
        self,
        # Launch parameters
        user_reference: str,
        project_id: str,
        pipeline_id: str,
        analysis_input: NextflowAnalysisInput,
        analysis_storage_id: Optional[str] = None,
        analysis_storage_size: Optional[AnalysisStorageSize] = None,
        activation_id: Optional[str] = None,
        # Output parameters
        output_parent_folder_id: Optional[str] = None,
        output_parent_folder_path: Optional[str] = None,
        analysis_output_uri: Optional[str] = None,
        # Meta parameters
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
        # CWL Specific parameters
        cwltool_overrides: Optional[Dict] = None
    ):
        """
        Initialise input
        :param user_reference
        :param project_id
        :param pipeline_id
        :param analysis_input
        :param analysis_storage_id
        :param analysis_storage_size
        :param activation_id
        :param output_parent_folder_id
        :param output_parent_folder_path
        :param analysis_output_uri
        :param tags
        :param cwltool_overrides
        """
        # Initialise any cwl specific parameters first before calling the parent class
        # Set cwl specific inputs
        self.cwltool_overrides: Optional[Dict] = cwltool_overrides
        # Set under parent init script through set_engine_parameters
        self.engine_parameters: Optional[ICAv2NextflowEngineParameters] = None

        # Call parent class
        super().__init__(
            user_reference=user_reference,
            project_id=project_id,
            pipeline_id=pipeline_id,
            analysis_input=analysis_input,
            analysis_storage_id=analysis_storage_id,
            analysis_storage_size=analysis_storage_size,
            activation_id=activation_id,
            output_parent_folder_id=output_parent_folder_id,
            output_parent_folder_path=output_parent_folder_path,
            analysis_output_uri=analysis_output_uri,
            tags=tags
        )

    def set_engine_parameters(self):
        self.engine_parameters = ICAv2NextflowEngineParameters(
            project_id=self.project_id,
            pipeline_id=self.pipeline_id,
            output_parent_folder_id=self.output_parent_folder_id,
            output_parent_folder_path=self.output_parent_folder_path,
            analysis_output=self.analysis_output,
            analysis_input=self.analysis_input,
            tags=self.tags,
            analysis_storage_id=self.analysis_storage_id,
            analysis_storage_size=self.analysis_storage_size,
            activation_id=self.activation_id,
        )

    def create_analysis(self) -> CreateNextflowAnalysis:
        return CreateNextflowAnalysis(
            user_reference=self.user_reference,
            pipeline_id=self.pipeline_id,
            tags=self.engine_parameters.tags(),
            activation_code_detail_id=self.engine_parameters.activation_id,
            analysis_input=self.analysis_input,
            analysis_storage_id=self.engine_parameters.analysis_storage_id,
            output_parent_folder_id=self.engine_parameters.output_parent_folder_id,
            analysis_output=self.engine_parameters.analysis_output
        )

    def launch_analysis(self) -> Analysis:
        from ..functions.project_pipelines_functions import launch_nextflow_workflow
        return launch_nextflow_workflow(
            project_id=self.project_id,
            nextflow_analysis=self.analysis,
        )



#!/usr/bin/env python3

"""
Functions for project management
"""

# Libica imports
from libica.openapi.v2 import ApiClient, ApiException
from libica.openapi.v2.api.project_api import ProjectApi
from libica.openapi.v2.model.project_paged_list import ProjectPagedList

# Local imports
from ...utils.configuration import get_icav2_configuration
from ...utils.logger import get_logger
from ...utils.globals import LIBICAV2_DEFAULT_PAGE_SIZE

# Logger helpers
logger = get_logger()


def get_project_id_from_project_name(
    project_name: str
) -> str:
    """
    Given a project name return the id of the project
    Will raise an error if the project id cannot be found

    :param project_name: The name of the project

    :return: The id of the project
    :rtype: str

    :raises ValueError, ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project import get_project_id_from_project_name

        project_id = get_project_id_from_project_name("my_project")

        print(project_id)
        # "1234-5678-9012-3456"
    """

    # Configuration
    configuration = get_icav2_configuration()

    # Collect the projects
    # We assume that there aren't more than 1000 projects anyway
    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        api_instance = ProjectApi(api_client)

    # example passing only required values which don't have defaults set
    # and optional values
    try:
        # Retrieve a list of projects.
        api_response: ProjectPagedList = api_instance.get_projects(
            page_size=LIBICAV2_DEFAULT_PAGE_SIZE,
         )
    except ApiException as e:
        logger.error("Exception when calling ProjectApi->get_projects: %s\n" % e)
        raise ApiException

    try:
        return next(
            filter(
                lambda project_iter: project_iter.name == project_name,
                api_response.items
            )
        )
    except StopIteration:
        raise ValueError(f"Project with name {project_name} not found")




#!/usr/bin/env python

# Import everything
from .functions.project_pipelines_functions import (
    # Project Data functions
    get_project_pipeline_obj,
    get_project_pipeline_id_from_pipeline_code,
    get_default_analysis_storage_id_from_project_pipeline,
    get_project_pipeline_description_from_pipeline_id,
    get_analysis_storage_id_from_analysis_storage_size,
    get_activation_id,
    get_best_matching_entitlement_detail_for_cwl_analysis,
    get_best_matching_entitlement_detail_for_nextflow_analysis,
    create_cwl_input_json_analysis_obj,
    launch_cwl_workflow,
    launch_nextflow_workflow,
    get_project_pipeline_input_parameters,
    get_project_pipeline_configuration_parameters,
    convert_icav2_uris_to_data_ids_from_cwl_input_json
)

from .classes.analysis import (
    # Wrapica classes
    ICAv2AnalysisInput,
    ICAv2PipelineAnalysisTags,
    ICAv2EngineParameters
)

from .classes.cwl_analysis import (
    # Wrapica classes
    ICAv2CwlAnalysisJsonInput,
    ICAv2CWLEngineParameters,
    ICAv2CWLPipelineAnalysis
)

from .classes.nextflow_analysis import (
    # Wrapica classes
    ICAv2NextflowAnalysisInput,
    ICAv2NextflowEngineParameters,
    ICAv2NextflowPipelineAnalysis
)

__all__ = [
    # Functions
    'get_project_pipeline_obj',
    'get_project_pipeline_id_from_pipeline_code',
    'get_default_analysis_storage_id_from_project_pipeline',
    'get_project_pipeline_description_from_pipeline_id',
    'get_analysis_storage_id_from_analysis_storage_size',
    'get_activation_id',
    'get_best_matching_entitlement_detail_for_cwl_analysis',
    'get_best_matching_entitlement_detail_for_nextflow_analysis',
    'create_cwl_input_json_analysis_obj',
    'launch_cwl_workflow',
    'launch_nextflow_workflow',
    'convert_icav2_uris_to_data_ids_from_cwl_input_json',
    # classes
    'ICAv2AnalysisInput',
    'ICAv2PipelineAnalysisTags',
    'ICAv2EngineParameters',
    'ICAv2CwlAnalysisJsonInput',
    'ICAv2CWLEngineParameters',
    'ICAv2CWLPipelineAnalysis',
    'ICAv2NextflowAnalysisInput',
    'ICAv2NextflowEngineParameters',
    'ICAv2NextflowPipelineAnalysis'
]



from . import (
    enums,
    libica_models,
    libica_exceptions,
    project,
    project_data,
    project_pipelines,
    project_analysis,
    job
)


__all__ = [
    # submodules
    project,
    project_data,
    project_pipelines,
    project_analysis,
    # Utils
    enums,
    libica_models,
    libica_exceptions,
    job
]
